# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=unnecessary-pass
# pylint: disable-msg=protected-access
# pylint: disable-msg=broad-except
# pylint: disable-msg=import-error
"""Qiskit Metal unit tests analyses functionality."""

import unittest
from pathlib import Path
from typing import Union

import numpy as np

from qiskit_metal.designs.design_multiplanar import MultiPlanar
from qiskit_metal.qlibrary.qubits.transmon_concentric import TransmonConcentric
from qiskit_metal.qlibrary.qubits.transmon_pocket_6 import TransmonPocket6
from qiskit_metal.tests.assertions import AssertionsMixin
from qiskit_metal.tests.test_data.quad_coupler import QuadCoupler
from qiskit_metal.toolbox_metal import (about, bounds_for_path_and_poly_tables,
                                        math_and_overrides, parsing)
from qiskit_metal.toolbox_metal.bounds_for_path_and_poly_tables import \
    BoundsForPathAndPolyTables
from qiskit_metal.toolbox_metal.exceptions import (IncorrectQtException,
                                                   InputError,
                                                   QiskitMetalDesignError,
                                                   QiskitMetalExceptions,
                                                   QLibraryGUIException)
from qiskit_metal.toolbox_metal.layer_stack_handler import LayerStackHandler

TEST_DATA = Path(__file__).parent / "test_data"


class TestToolboxMetal(unittest.TestCase, AssertionsMixin):
    """Unit test class."""

    def setUp(self):
        """Setup unit test."""
        pass

    def tearDown(self):
        """Tie any loose ends."""
        math_and_overrides.set_decimal_precision(10)

    def test_toolbox_metal_instantiation_qiskit_metal_exceptions(self):
        """Test instantiation of QiskitMetalExceptions."""
        try:
            QiskitMetalExceptions("test message")
        except Exception:
            self.fail("QiskitMetalExceptions failed.")

    def test_toolbox_metal_instantiation_qiskit_metal_design_error(self):
        """Test instantiation of QiskitMetalDesignError."""
        try:
            QiskitMetalDesignError("test message")
        except Exception:
            self.fail("QiskitMetalDesignError failed.")

    def test_toolbox_metal_instantiation_incorrect_qt_exception(self):
        """Test instantiation of IncorrectQtException."""
        try:
            IncorrectQtException("test message")
        except Exception:
            self.fail("IncorrectQtException failed.")

    def test_toolbox_metal_instantiation_qlibrary_gui_exception(self):
        """Test instantiation of QLibraryGUIException."""
        try:
            QLibraryGUIException("test message")
        except Exception:
            self.fail("QLibraryGUIException failed.")

    def test_toolbox_metal_instantiation_input_error(self):
        """Test instantiation of InputError."""
        try:
            InputError("test message")
        except Exception:
            self.fail("InputError failed.")

    def test_toolbox_metal_about(self):
        """Test that about in about.py produces about text without any
        errors."""
        try:
            about.about()
        except Exception:
            self.fail("about() failed")

    def test_toolbox_metal_open_docs(self):
        """Test that open_docs in about.py opens qiskit_metal documentation in HTML without any
        errors."""
        try:
            about.open_docs()
        except Exception:
            self.fail("open_docs() failed")

#    def test_toolbox_metal_orient_me(self):
        """Test that orient_me in about.py produces detailed information about the user without any
        errors."""


#        try:
#            about.orient_me(True)
#        except Exception:
#            self.fail("orient_me() failed")
#        try:
#            about.orient_me(False)
#        except Exception:
#            self.fail("orient_me() failed")

    def test_toolbox_metal_get_platform_info(self):
        """Test that get_platform_info in about.py returns a string with the platform information without any errors."""
        try:
            return about.get_platform_info()
        except Exception:
            self.fail("get_platform_info() failed")

    # pylint: disable-msg=unused-variable
    def test_toolbox_metal_parsing_true_str(self):
        """Test that TRUE_STR in parsing.py has not accidentally changed."""
        expected = [
            'True', 'true', 'TRUE', True, '1', 't', 'y', 'Y', 'YES', 'yes',
            'yeah', 1, 1.0
        ]
        actual = parsing.TRUE_STR

        self.assertEqual(len(actual), len(expected))

        for x, _ in enumerate(expected):
            self.assertTrue(_ in actual)

    def test_toolbox_metal_is_true(self):
        """Test is_true in toolbox_metal.py."""
        self.assertTrue(parsing.is_true('true'))
        self.assertTrue(parsing.is_true('True'))
        self.assertTrue(parsing.is_true('TRUE'))
        self.assertTrue(parsing.is_true('1'))
        self.assertTrue(parsing.is_true('t'))
        self.assertTrue(parsing.is_true('y'))
        self.assertTrue(parsing.is_true('Y'))
        self.assertTrue(parsing.is_true('YES'))
        self.assertTrue(parsing.is_true('yes'))
        self.assertTrue(parsing.is_true('yeah'))
        self.assertTrue(parsing.is_true(True))
        self.assertTrue(parsing.is_true(1))
        self.assertTrue(parsing.is_true(1.0))

        self.assertFalse(parsing.is_true('false'))
        self.assertFalse(parsing.is_true('False'))
        self.assertFalse(parsing.is_true('FALSE'))
        self.assertFalse(parsing.is_true('0'))
        self.assertFalse(parsing.is_true('f'))
        self.assertFalse(parsing.is_true(False))
        self.assertFalse(parsing.is_true(78))
        self.assertFalse(parsing.is_true('gibberish'))
        self.assertFalse(parsing.is_true({'key': 'val'}))

    def test_toolbox_metal_parse_string_to_float(self):
        """Test _parse_string_to_float in toolbox_metal.py."""
        self.assertAlmostEqualRel(parsing._parse_string_to_float('1um'),
                                  0.001,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing._parse_string_to_float('2mm'),
                                  2,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing._parse_string_to_float('3m'),
                                  3000.0,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing._parse_string_to_float('4km'),
                                  4000000.0,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing._parse_string_to_float('5cm'),
                                  50.0,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing._parse_string_to_float('6mm'),
                                  6,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing._parse_string_to_float('2 mile'),
                                  3218688.0,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing._parse_string_to_float('3 yard'),
                                  2743.2,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing._parse_string_to_float('4 inch'),
                                  101.6,
                                  rel_tol=1e-3)

        self.assertEqual(parsing._parse_string_to_float('not-a-number'),
                         'not-a-number')
        self.assertEqual(parsing._parse_string_to_float(12), 12)
        self.assertEqual(parsing._parse_string_to_float('12.2.3'), '12.2.3')

    def test_toolbox_metal_is_variable_name(self):
        """Test is_variable_name in toolbox_metal.py."""
        self.assertTrue(parsing.is_variable_name('ok'))
        self.assertTrue(parsing.is_variable_name('still_ok'))
        self.assertTrue(parsing.is_variable_name('EXPAT'))
        self.assertTrue(parsing.is_variable_name('for'))
        self.assertTrue(parsing.is_variable_name('CamelCase'))

        self.assertFalse(parsing.is_variable_name('1_fix'))
        self.assertFalse(parsing.is_variable_name('12'))

    def test_toolbox_metal_is_for_ast_eval(self):
        """Test is_for_ast_eval in toolbox_metal.py."""
        self.assertTrue(parsing.is_for_ast_eval('[1, 2, 3]'))
        self.assertTrue(parsing.is_for_ast_eval('[[1, 2], [3, 4]]'))
        self.assertTrue(parsing.is_for_ast_eval('{1, 2}'))
        self.assertTrue(parsing.is_for_ast_eval('[{1, 2}]'))
        self.assertTrue(parsing.is_for_ast_eval('[{1, 2}, [3, 4]]'))
        self.assertTrue(parsing.is_for_ast_eval('[{1, 2}, {3, 4}]'))

        self.assertFalse(parsing.is_for_ast_eval('text'))
        self.assertFalse(parsing.is_for_ast_eval('12'))
        self.assertFalse(parsing.is_for_ast_eval('[1, 2}'))

    def test_toolbox_metal_is_numeric_possible(self):
        """Test is_numeric_possible in toolbox_metal.py."""
        self.assertTrue(parsing.is_numeric_possible('123'))
        self.assertTrue(parsing.is_numeric_possible('12mm'))
        self.assertTrue(parsing.is_numeric_possible('123.456'))
        self.assertTrue(parsing.is_numeric_possible('127.0.0.1'))
        self.assertTrue(parsing.is_numeric_possible('12.2um'))
        self.assertTrue(parsing.is_numeric_possible('-8um'))
        self.assertTrue(parsing.is_numeric_possible('+8um'))

        self.assertFalse(parsing.is_numeric_possible('abc'))
        self.assertFalse(parsing.is_numeric_possible('mm12'))
        self.assertFalse(parsing.is_numeric_possible('explode'))

    def test_toolbox_metal_parse_value(self):
        """Test parse_value in toolbox_metal.py."""
        var_dict = {'data_a': '4 miles', 'data_b': '1um'}

        self.assertAlmostEqualRel(parsing.parse_value('1 meter', var_dict),
                                  1000.0,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing.parse_value('2 um', var_dict),
                                  0.002,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing.parse_value('3 miles', var_dict),
                                  4828032.0,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing.parse_value('4 feet', var_dict),
                                  1219.1999999999998,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing.parse_value('5 km', var_dict),
                                  5000000.0,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing.parse_value('6 pm', var_dict),
                                  6.000000000000001e-09,
                                  rel_tol=1e-10)
        self.assertAlmostEqualRel(parsing.parse_value('-12 inches', var_dict),
                                  -304.79999999999995,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing.parse_value('data_a', var_dict),
                                  6437376.0,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing.parse_value('data_b', var_dict),
                                  0.001,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing.parse_value('1 um', None),
                                  0.001,
                                  rel_tol=1e-3)
        self.assertAlmostEqualRel(parsing.parse_value('1 um', 'None'),
                                  0.001,
                                  rel_tol=1e-3)

    def test_toolbox_metal_parse_options(self):
        """Test parse_options in toolbox_metal.py."""
        dict_1 = {'data_a': '2mm', 'data_b': '1um'}
        dict_2 = {
            'data_a': '20mm',
            'data_b': '15um',
            'data_c': '4 miles',
            'data_d': '2 mm'
        }

        expected = [[2], [0.001], [20, 0.015, 6437376.0, 2]]

        actual = []
        actual.append(parsing.parse_options(dict_1, 'data_a'))
        actual.append(parsing.parse_options(dict_1, 'data_b'))
        actual.append(
            parsing.parse_options(dict_2, 'data_a,data_b,data_c,data_d',
                                  dict_1))

        for x in range(3):
            self.assertEqual(len(actual[x]), len(expected[x]))
            my_range = len(actual[x])
            for i in range(my_range):
                self.assertAlmostEqualRel(actual[x][i],
                                          expected[x][i],
                                          rel_tol=1e-3)

    def test_toolbox_metal_extract_value_unit(self):
        """Test functionality of extract_value_unit in toolbox_metal.py."""
        self.assertEqual(parsing.extract_value_unit("200mm", "mm"), 200.0)
        self.assertEqual(parsing.extract_value_unit("20.5", "units"), 20.5)

    def test_toolbox_metal_fix_units(self):
        """Test functionality of fix_units in toolbox_metal.py."""
        self.assertIsInstance(parsing.fix_units(2.5, "mm"), str)
        self.assertEqual(parsing.fix_units(10.5, None), "10.5mm")

    def test_toolbox_metal_parse_entry(self):
        """Test functionality of parse_entry in toolbox_metal.py."""
        self.assertIsInstance(parsing.parse_entry([2, "two"]), list)
        self.assertEqual(parsing.parse_entry(5.5), 5.5)
        self.assertIsInstance(parsing.parse_entry("test"), str)

    def test_toolbox_metal_parse_units(self):
        """Test functionality of parse_units in toolbox_metal.py."""
        self.assertEqual(parsing.parse_units(8), 0.008)
        self.assertEqual(parsing.parse_units("two"), "two")
        self.assertEqual(parsing.parse_units([2, 3.5, 5]),
                         [0.002, 0.0035, 0.005])

    def test_toolbox_metal_set_decimal_precision(self):
        """Test functionality of set_decimal_precision in toolbox_metal.py."""
        self.assertEqual(math_and_overrides.DECIMAL_PRECISION, 10)
        math_and_overrides.set_decimal_precision(15)
        self.assertEqual(math_and_overrides.DECIMAL_PRECISION, 15)

    def test_toolbox_metal_dot(self):
        """Test functionality of dot in toolbox_metal.py."""
        math_and_overrides.set_decimal_precision(3)
        my_array_1 = np.array([3, 4])
        my_array_2 = np.array([12, 14])
        self.assertEqual(math_and_overrides.dot(my_array_1, my_array_2), 92)

    def test_toolbox_metal_round(self):
        """Test functionality of round in toolbox_metal.py."""
        math_and_overrides.set_decimal_precision(2)
        self.assertEqual(math_and_overrides.round(12.349), 12.35)

    def test_toolbox_metal_cross(self):
        """Test functionality of cross in toolbox_metal.py."""
        math_and_overrides.set_decimal_precision(3)
        my_array_1 = np.array([3, 4])
        my_array_2 = np.array([12, 14])
        self.assertEqual(math_and_overrides.cross(my_array_1, my_array_2), -6)

    def test_toolbox_metal_aligned_pts(self):
        """Test functionality of aligned_pts in toolbox_metal.py."""
        self.assertTrue(math_and_overrides.aligned_pts([0, 1, 2]))
        self.assertFalse(math_and_overrides.aligned_pts([7, 1, 2]))

    def test_toolbox_metal_determine_larger_box(self):
        """Test functionality of determine_larger_box in toolbox_metal.py."""
        self.assertEqual(
            bounds_for_path_and_poly_tables.determine_larger_box(
                0.0, 0.0, 11.0, 11.0, (0.0, 0.0, 10.0, 10.0)),
            (0.0, 0.0, 10.0, 10.0))

    def test_toolbox_metal_get_bounds_of_path_and_poly_tables(self):
        """Test functionality of get_bounds_of_path_and_poly_tables in toolbox_metal.py"""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))

        self.assertIsInstance(
            BoundsForPathAndPolyTables(
                multiplanar_design).get_bounds_of_path_and_poly_tables(
                    False, [], 1, 0, 0), tuple)

        self.assertEqual(
            len(
                BoundsForPathAndPolyTables(
                    multiplanar_design).get_bounds_of_path_and_poly_tables(
                        False, [], 1, 0, 0)), 5)

    def test_toolbox_metal_ensure_component_box_smaller_than_chip_box_(self):
        """Test functionality of ensure_component_box_smaller_than_chip_box in toolbox_metal.py"""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        chip = getattr(multiplanar_design, "_chips")
        for main in chip.items():
            coords = main[1]
        for item in coords.values():
            x_0 = item['center_x']
            y_0 = item['center_y']
            x_1 = item['size_x']
            y_1 = item['size_y']
        x_0 = int(float(x_0[:-2]))
        y_0 = int(float(y_0[:-2]))
        x_1 = int(x_1[:-2])
        y_1 = int(y_1[:-2])
        self.assertEqual(
            BoundsForPathAndPolyTables(
                multiplanar_design).ensure_component_box_smaller_than_chip_box_(
                    (0, 0, 2, 3), (x_0, y_0, x_1, y_1)), (0, 0, 2, 3))
        self.assertEqual(
            BoundsForPathAndPolyTables(
                multiplanar_design).ensure_component_box_smaller_than_chip_box_(
                    (0, 0, 9, 7), (None, None, None, None)), (0, 0, 9, 7))

    def test_toolbox_metal_get_box_for_xy_bounds(self):
        """Test functionality of get_box_for_xy_bounds in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        multiplanar_design.chips['c_chip'] = {
            'size': {
                'center_x': '0.0mm',
                'center_y': '0.0mm',
                'size_x': '9mm',
                'size_y': '7mm'
            }
        }
        self.assertEqual(
            BoundsForPathAndPolyTables(
                multiplanar_design).get_box_for_xy_bounds(),
            (-4.5, -3.5, 4.5, 3.5))

    def test_toolbox_metal_are_all_chipnames_in_design(self):
        """Test functionality of are_all_chipnames_in_design in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        multiplanar_design.chips['c_chip'] = {
            'size': {
                'center_x': '0.0mm',
                'center_y': '0.0mm',
                'size_x': '9mm',
                'size_y': '7mm'
            }
        }
        self.assertEqual(
            BoundsForPathAndPolyTables(
                multiplanar_design).are_all_chipnames_in_design(),
            (True, {'c_chip'}))

    def test_toolbox_metal_get_x_y_for_chip(self):
        """Test functionality of get_x_y_for_chip in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        multiplanar_design.chips['c_chip'] = {
            'size': {
                'center_x': '0.0mm',
                'center_y': '0.0mm',
                'size_x': '9mm',
                'size_y': '7mm'
            }
        }
        self.assertEqual(
            BoundsForPathAndPolyTables(multiplanar_design).get_x_y_for_chip(
                'c_chip'), ((-4.5, -3.5, 4.5, 3.5), 0))
        self.assertEqual(
            BoundsForPathAndPolyTables(multiplanar_design).get_x_y_for_chip(
                'TEST_chip'), ((), 1))

    def test_toolbox_metal_chip_names_not_in_design(self):
        """Test functionality of chip_names_not_in_design in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            BoundsForPathAndPolyTables(
                multiplanar_design).chip_names_not_in_design(
                    set(multiplanar_design.chips.keys()),
                    multiplanar_design.ls.get_unique_chip_names()), None)

    def test_toolbox_metal_chip_size_not_in_chipname_within_design(self):
        """Test functionality of chip_size_not_in_chipname_within_design in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            BoundsForPathAndPolyTables(multiplanar_design).
            chip_size_not_in_chipname_within_design('c_chip'), None)

    def test_toolbox_metal_get_layer_datatype_when_fill_is_true(self):
        """Test functionality of get_layer_datatype_when_fill_is_true in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            LayerStackHandler(
                multiplanar_design,
                (ls_file_path, None)).get_layer_datatype_when_fill_is_true(), {
                    (1, 0): {
                        'layer': 1,
                        'datatype': 0,
                        'thickness': 0.002,
                        'z_coord': 0.01,
                        'material': 'pec',
                        'chip_name': 'c_chip'
                    },
                    (3, 0): {
                        'layer': 3,
                        'datatype': 0,
                        'thickness': -0.5,
                        'z_coord': 0.0,
                        'material': 'silicon',
                        'chip_name': 'c_chip'
                    }
                })

    def test_toolbox_metal_get_properties_for_layer_datatype(self):
        """Test functionality of get_properties_for_layer_datatype in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            LayerStackHandler(
                multiplanar_design,
                (ls_file_path, None)).get_properties_for_layer_datatype(
                    ['material', 'thickness'], 1, 0), ('pec', 0.002))
        self.assertEqual(
            LayerStackHandler(
                multiplanar_design,
                (ls_file_path, None)).get_properties_for_layer_datatype(
                    ['material', 'thickness'], 3, 0), ('silicon', -0.5))

    def test_toolbox_metal_is_layer_data_unique(self):
        """Test functionality of is_layer_data_unique in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            LayerStackHandler(multiplanar_design,
                              (ls_file_path, None)).is_layer_data_unique(),
            True)

    def test_toolbox_metal_read_csv_df(self):
        """Test functionality of read_csv_df in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            LayerStackHandler(multiplanar_design,
                              (ls_file_path, None))._read_csv_df(ls_file_path),
            None)

    def test_toolbox_metal_get_unique_chip_names(self):
        """Test functionality of get_unique_chip_names in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            LayerStackHandler(multiplanar_design,
                              (ls_file_path, None)).get_unique_chip_names(),
            {'c_chip'})

    def test_toolbox_metal_get_unique_layer_ints(self):
        """Test functionality of get_unique_layer_ints in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            LayerStackHandler(multiplanar_design,
                              (ls_file_path, None)).get_unique_layer_ints(),
            {1, 3})

    def test_toolbox_metal_warning_properties(self):
        """Test functionality of _warning_properties in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            LayerStackHandler(multiplanar_design,
                              (ls_file_path, None))._warning_properties(
                                  ['test']), None)

    def test_toolbox_metal_warning_search(self):
        """Test functionality of _warning_search in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            LayerStackHandler(multiplanar_design,
                              (ls_file_path, None))._warning_search(
                                  'test_chip', 7, 1, "FAIL"), None)

    def test_toolbox_metal_warning_search_minus_chip(self):
        """Test functionality of _warning_search_minus_chip in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            LayerStackHandler(multiplanar_design,
                              (ls_file_path, None))._warning_search_minus_chip(
                                  5, 1, "FAIL"), None)

    def test_toolbox_metal_layer_stack_handler_pilot_error(self):
        """Test functionality of layer_stack_handler_pilot_error in toolbox_metal.py."""
        ls_file_path = TEST_DATA / 'planar_chip.txt'
        multiplanar_design = MultiPlanar(metadata={},
                                         overwrite_enabled=True,
                                         layer_stack_filename=ls_file_path)
        conn_pads = dict(connection_pads=dict(coup1=dict(loc_W=-1, loc_H=1),
                                              coup2=dict(loc_W=1, loc_H=1)))
        q1 = TransmonPocket6(multiplanar_design,
                             "Q1",
                             options=dict(**conn_pads, chip="c_chip", layer=1))
        qc1 = QuadCoupler(multiplanar_design,
                          "qc1",
                          options=dict(pos_x="0mm",
                                       pos_y="-0.08mm",
                                       pad_width="120um",
                                       pad_height="30um",
                                       cpw_stub_height="250um",
                                       chip="c_chip",
                                       layer=1))
        self.assertEqual(
            LayerStackHandler(
                multiplanar_design,
                (ls_file_path, None)).layer_stack_handler_pilot_error(), None)

if __name__ == '__main__':
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
=================================================
Analyses (:mod:`qiskit_metal.analyses`)
=================================================

.. currentmodule:: qiskit_metal.analyses

Module containing all Qiskit Metal analyses.

.. _qanalysis:


Analysis Core
-------------

.. autosummary::
    :toctree:

    QAnalysis
    QSimulation

Hamiltonian
-----------

.. autosummary::
    :toctree:

    Hcpb
    Hcpb_analytic
    HO_wavefunctions
    transmon_analytics

Electromagnetic & quantization / parameter extraction
-----------------------------------------------------

.. autosummary::
    :toctree:

    cpw_calculations
    kappa_calculation

Sweeping Options
----------------

.. autosummary::
    :toctree:

    Sweeper

Quantization
------------

.. autosummary::
    :toctree:

    LOManalysis
    EPRanalysis
    lumped_capacitive

Simulation
------------

.. autosummary::
    :toctree:

    LumpedElementsSim
    EigenmodeSim
    ScatteringImpedanceSim

"""

from .. import config

from .core import QAnalysis
from .core import QSimulation
from .em import cpw_calculations
from .em import kappa_calculation
from .quantization import lumped_capacitive
from .quantization import EPRanalysis
from .quantization import LOManalysis
from .simulation import LumpedElementsSim
from .simulation import EigenmodeSim
from .simulation import ScatteringImpedanceSim
from .hamiltonian.transmon_charge_basis import Hcpb
from .hamiltonian import HO_wavefunctions
from .hamiltonian import transmon_analytics
from .hamiltonian.transmon_CPB_analytic import Hcpb_analytic
from .sweep_and_optimize.sweeper import Sweeper



#
"""
Variables

from pyaedt.HFSS import HFSS
with HFSS as hfss:
     hfss["dim"] = "1mm"   # design variable
     hfss["$dim"] = "1mm"  # project variable

Modeler

# Create a box, assign variables, and assign materials.

from pyaedt.hfss import Hfss
with Hfss as hfss:
     hfss.modeler.create_box([0, 0, 0], [10, "dim", 10],
                             "mybox", "aluminum")


"""

from qiskit_metal.renderers.renderer_ansys_pyaedt.pyaedt_base import QPyaedt
from qiskit_metal.renderers.renderer_ansys.ansys_renderer import QAnsysRenderer
from qiskit_metal.draw.utility import to_vec3D, to_vec3D_list
from qiskit_metal.toolbox_metal.parsing import parse_entry, parse_units

from qiskit_metal import Dict
from typing import List, Tuple, Union
import pandas as pd
import shapely

from qiskit_metal import config
if not config.is_building_docs():
    from qiskit_metal.toolbox_python.utility_functions import get_clean_name


class QHFSSPyaedt(QPyaedt):
    """Subclass of pyaedt renderer for running HFSS simulations.

     QPyaedt Default Options:

    """

    default_setup = Dict(
        drivenmodal=Dict(
            name="Setup",
            freq_ghz="5.0",
            max_delta_s="0.1",
            max_passes="10",
            min_passes="1",
            min_converged="1",
            pct_refinement="30",
            basis_order="1",
        ),
        eigenmode=Dict(
            name="Setup",
            min_freq_ghz="1",
            n_modes="1",
            max_delta_f="0.5",
            max_passes="10",
            min_passes="1",
            min_converged="1",
            pct_refinement="30",
            basis_order="-1",
        ),
    ),
    name = 'aedt_hfss'

    aedt_hfss_options = Dict(
        # spacing between port and inductor if junction is drawn both ways
        port_inductor_gap='10um')
    """aedt HFSS Options"""

    def __init__(self,
                 multilayer_design: 'MultiPlanar',
                 renderer_type: str = 'HFSS',
                 project_name: Union[str, None] = None,
                 design_name: Union[str, None] = None,
                 initiate=False,
                 options: Dict = None):
        """Create a QRenderer for HFSS simulations using pyaedt and multiplanar design.
        QHFSSPyaedt is subclassed from QPyaedt, subclassed from QRendererAnalysis and
        subclassed from QRenderer.  The default_setup options are expected to be defined by
        child class of QHFSSPyaedt for driven-modal and eigenmode solution types.


        Args:

            multilayer_design (MultiPlanar): Use QGeometry within MultiPlanar to obtain elements for Ansys.
            renderer_type (str): Choose string from list ['HFSS_DM', 'HFSS_EM', 'Q3D'] for the type of
                                design to insert into project.
            project_name (Union[str, None], optional): Give a name, or one will be made based on class name of renderer.
                                            Defaults to None.
            design_name (Union[str, None], optional): Give a name, or one will be made based on class name of renderer.
                                            Defaults to None.
            initiate (bool, optional): True to initiate the renderer. Defaults to False.
            options (Dict, optional):  Used to override all options. Defaults to None.
        """

        super().__init__(multilayer_design,
                         renderer_type=renderer_type,
                         project_name=project_name,
                         design_name=design_name,
                         initiate=initiate,
                         options=options)

        self.current_sweep = None

        # Check if user entered valid data
        self.port_list_is_valid = None  # bool
        self.jj_to_port_is_valid = None  #bool
        self.ignored_jjs_is_valid = None  #bool

        # Reformat the port_list and enter into port_list_dict
        self.port_list_dict = Dict()

        # Reformat the jj_to_port and enter int jj_to_port_dict
        self.jj_to_port_dict = Dict()

        # Keep copy of ignored_jjs to use after rendering.
        self.ignored_jjs = []

        # QRenderer has a "cls" method called load()
        if_registered_in_design = QHFSSPyaedt.load()

        #make a class to read in pandas table.
        self.tables = None

    def render_design(self,
                      selection: Union[list, None] = None,
                      open_pins: Union[list, None] = None,
                      port_list: Union[list, None] = None,
                      jj_to_port: Union[list, None] = None,
                      ignored_jjs: Union[list, None] = None,
                      box_plus_buffer: bool = True):
        """
        This render_design will add additional logic for HFSS within project.
        In particular, when used by BOTH solution types of eigenmode and drivenmodal.

        Initiate rendering of components in design contained in selection,
        assuming they're valid. Components are rendered before the chips they
        reside on, and subtraction of negative shapes is performed at the very
        end.

        First obtain a list of IDs of components to render and a corresponding case, denoted by self.qcomp_ids
        and self.case, respectively. If self.case == 1, all components in QDesign are to be rendered.
        If self.case == 0, a strict subset of components in QDesign are to be rendered. Otherwise, if
        self.case == 2, one or more component names in selection cannot be found in QDesign.

        Chip_subtract_dict consists of component names (keys) and a set of all elements within each component that
        will eventually be subtracted from the ground plane. Add objects that are perfect conductors and/or have
        meshing to self.assign_perfE and self.assign_mesh, respectively; both are initialized as empty lists. Note
        that these objects are "refreshed" each time render_design is called (as opposed to in the init function)
        to clear QAnsysRenderer of any leftover items from the last call to render_design.

        Among the components selected for export, there may or may not be unused (unconnected) pins.
        The second parameter, open_pins, contains tuples of the form (component_name, pin_name) that
        specify exactly which pins should be open rather than shorted during the simulation. Both the
        component and pin name must be specified because the latter could be shared by multiple
        components. All pins in this list are rendered with an additional endcap in the form of a
        rectangular cutout, to be subtracted from its respective plane.

        In driven modal solutions, the Ansys design must include one or more
        ports. This is done by adding all port designations and their
        respective impedances in Ohms as (qcomp, pin, impedance) to port_list.
        Note that an open endcap must separate the two sides of each pin before
        inserting a lumped port in between, so behind the scenes all pins in
        port_list are also added to open_pins. Practically, however, port_list
        and open_pins are inputted as mutually exclusive lists.

        Also in driven modal solutions, one may want to render junctions as
        lumped ports and/or inductors, or omit them altogether. To do so,
        tuples of the form (component_name, element_name, impedance)
        are added to the list jj_to_port. For example,
        ('Q1', 'rect_jj', 70) indicates that rect_jj of component Q1 is
        to be rendered as both a lumped port with an impedance of 70 Ohms.

        Alternatively for driven modal solutions, one may want to disregard
        select junctions in the Metal design altogether to simulate the
        capacitive effect while keeping the qubit in an "off" state. Such
        junctions are specified in the form (component_name, element_name)
        in the list ignored_jjs.


        The final parameter, box_plus_buffer, determines how the chip is drawn. When set to True, it takes the
        minimum rectangular bounding box of all rendered components and adds a buffer of x_buffer_width_mm and
        y_buffer_width_mm horizontally and vertically, respectively, to the chip size. The center of the chip
        lies at the midpoint x/y coordinates of the minimum rectangular bounding box and may change depending
        on which components are rendered and how they're positioned. If box_plus_buffer is False, however, the
        chip position and dimensions are taken from the chip info dictionary found in self.design, irrespective
        of what's being rendered. While this latter option is faster because it doesn't require calculating a
        bounding box, it runs the risk of rendered components being too close to the edge of the chip or even
        falling outside its boundaries.

        Args:
            selection (Union[list, None], optional): List of components to render. Defaults to None.
            open_pins (Union[list, None], optional): List of tuples of pins that are open. Defaults to None.
            port_list (Union[list, None], optional): List of tuples of pins to
                                        be rendered as ports. Defaults to None.
            jj_to_port (Union[list, None], optional): List of tuples of jj's to
                                        be rendered as ports. Defaults to None.
            ignored_jjs (Union[list, None], optional): List of tuples of jj's
                                        that shouldn't be rendered.
                                        Defaults to None.
            box_plus_buffer (bool, optional): Either calculate a bounding box based on the
                                            location of rendered geometries or use chip size
                                            from design class.
        """
        # Draw in fill = True pieces. based on either full chip or box_plus

        # They are reset for each time render_design happens.
        super().render_design(selection, box_plus_buffer)

        if self.case == 2:
            self.logger.warning(
                'Unable to proceed with rendering. Please check selection.')
            return

        self.reset_hfss_arguments()
        self.ignored_jjs = ignored_jjs

        #Every time one chooses to render, we have to get the chip names since the
        #qgeometry tables can be updated between each render_design().
        chip_list = self.get_chip_names()

        if not self.valid_input_arguments(open_pins, port_list, jj_to_port,
                                          ignored_jjs):
            self.logger.error(
                f'Check the arguments to render_design, invalid name was probably used.'
            )
        else:
            # Ansys default units is 'mm'?????, but metal using 'meter'.
            # pyaedt.generic.constants.SI_UNITS.Length is meter.
            self.activate_user_project_design()

            # Add something to clear design.

            self.draw_sample_holder()

            self.aedt_render_by_layer_then_tables(skip_junction=False,
                                                  open_pins=open_pins,
                                                  port_list=port_list,
                                                  jj_to_port=jj_to_port,
                                                  ignored_jjs=ignored_jjs)

        # self.assign_thin_conductor()
        # self.assign_nets()

    def aedt_render_by_layer_then_tables(self,
                                         open_pins: Union[list, None],
                                         port_list: Union[list, None],
                                         jj_to_port: Union[list, None],
                                         ignored_jjs: Union[list, None],
                                         skip_junction: bool = False,
                                         data_type: int = 0):
        """_summary_

        Args:
            skip_junction (bool, optional): Should the renderer add junctions to HFSS.
                                            Defaults to False.
            open_pins (Union[list, None], optional): Which pins will have small rectangle of substrate
                                subtracted to be an endcap. Use definition from render_design.
                                Defaults to None.
            port_list (Union[list, None], optional): Use definition from render_design.
                                                    Defaults to None.
            jj_to_port (Union[list, None], optional): Use definition from render_design.
                                                    Defaults to None.
            ignored_jjs (Union[list, None], optional): Use definition from render_design.
                                                    Defaults to None.
            datatype (int, optional): This comes for the layer_stack file; regarding
                                                    layer and datatype. Defaults to 0.
        """

        super().aedt_render_by_layer_then_tables(skip_junction=skip_junction,
                                                 open_pins=open_pins,
                                                 port_list=port_list,
                                                 jj_to_port=jj_to_port,
                                                 ignored_jjs=ignored_jjs,
                                                 data_type=data_type)

        for layer_num in sorted(
                self.design.qgeometry.get_all_unique_layers_for_all_tables(
                    self.qcomp_ids)):

            # Create subsets per layer for  port_list, jj_to)port, and open_pins.

            result_subset = self.get_subset_based_on_layer(
                open_pins, port_list, jj_to_port, ignored_jjs, layer_num)
            open_pins_subset, port_list_subset, jj_to_port_subset, ignored_jjs_subset = result_subset

            for table_type in self.design.qgeometry.get_element_types():
                if table_type != 'junction' or not skip_junction:
                    #At this point, we know table type,
                    self.render_components(table_type, layer_num,
                                           port_list_subset, jj_to_port_subset,
                                           ignored_jjs_subset)

            #For each layer, add the endcaps and
            # add polyline to subtract to self.chip_subtract_dict[layer_num].

            if isinstance(open_pins_subset, list) and self.open_pins_is_valid:
                # Only use this method if user defines open pins.
                if open_pins_subset:
                    #Confirm there is something in the list.
                    self.add_endcaps(open_pins_subset, layer_num)

            solution_type = self.current_app.solution_type

            if solution_type == 'Modal' and self.port_list_is_valid and isinstance(
                    port_list_subset, list):

                # Only use this method if user defines drivenmodal solution_type,
                # and  port list is provided by user.
                if port_list_subset:
                    #Confirm there is something in the list.
                    self.port_list_dict_populate(port_list_subset)
                    self.add_ports(port_list_subset, self.port_list_dict,
                                   layer_num)

            # The layer is obtained from qgeometry tables, but layer_stack
            # can have multiple data_types.  For now, the subtract will
            # happen from data_type==0.
            self.subtract_from_ground(layer_num, data_type=data_type)

#yapf: disable
    def get_subset_based_on_layer(self,
                    open_pins: Union[list, None],
                    port_list: Union[list, None],
                    jj_to_port: Union[list, None],
                    ignored_jjs: Union[list,None],
                    layer_num: int) -> Tuple[
                                            Union[list, None],
                                            Union[list, None],
                                            Union[list, None],
                                            Union[list, None]]:
        #yapf: enable
        """When user adds arguments through render_design, they do not identify the layer that the
        component is on.  So This method returns a subset of the components which are to be
        rendered for the layer_num passed to this method.

        Args:
            open_pins (Union[list, None]):  Use definition from render_design.
            port_list (Union[list, None]):  Use definition from render_design.
            jj_to_port (Union[list, None]): Use definition from render_design.
            ignored_jjs (Union[list,None]):  Use definition from render_design.
            layer_num (int): Using this layer number, identify the components
                        in the remaining arguments

        Returns:
            Tuple[ Union[list, None], Union[list, None], Union[list, None], Union[list, None]]: Subset
            of the components which are to be rendered for the layer_num passed to this method.
        """
        open_pins_subset = None
        port_list_subset = None
        jj_to_port_subset = None
        ignored_jjs_subset = None

        if open_pins:
            open_pins_subset = self.get_open_pins_in_layer(open_pins, layer_num)

        if port_list:
            port_list_subset = self.get_port_list_in_layer(port_list, layer_num)

        if jj_to_port:
            jj_to_port_subset = self.get_jj_to_port_in_layer(
                jj_to_port, layer_num)

        if ignored_jjs:
            ignored_jjs_subset = self.get_ignored_jjs_in_layer(
                ignored_jjs, layer_num)

        return open_pins_subset, port_list_subset, jj_to_port_subset, ignored_jjs_subset

    def get_port_list_in_layer(self, port_list: list, layer_num: int) -> list:
        """Reduce the port_list list and only return the port_list for layer_nums.

        Args:
            port_list (list): The list given by user for render_design.
            layer_num (int): The layer number being rendered.

        Returns:
            list: A subset of port_list for the denoted layer_num.
        """
        layer_port_list = list()

        mask_comp_element = self.path_poly_and_junction_with_valid_comps[
            'layer'] == layer_num
        path_poly_junction_layer = self.path_poly_and_junction_with_valid_comps[
            mask_comp_element]

        for comp_name, element, impedance in port_list:
            id = self.design.components[comp_name].id
            comp_element_layer_mask = (
                path_poly_junction_layer['component'] == id)
            valid_df = path_poly_junction_layer[comp_element_layer_mask]
            if not valid_df.empty:
                layer_port_list.append((comp_name, element, impedance))

        return layer_port_list

    def get_jj_to_port_in_layer(self, jj_to_port: list, layer_num: int) -> list:
        """Reduce the port_list list and only return the port_list for layer_nums.
        This is very similar to get_port_list_in_layer.  This format is different
        than what was done with renderer with comm port. The inductance and creation
        of extra sheet has been dropped.

        Args:
            jj_to_port (list): Use definition from render_design.
            layer_num (int): The layer number to render.

        Returns:
            list: Use definition from render_design.
        """
        layer_jj_to_port = list()
        mask_comp_element = self.path_poly_and_junction_with_valid_comps[
            'layer'] == layer_num
        path_poly_junction_layer = self.path_poly_and_junction_with_valid_comps[
            mask_comp_element]

        for comp_name, element, impedance in jj_to_port:
            id = self.design.components[comp_name].id
            comp_element_layer_mask = (
                path_poly_junction_layer['component'] == id)
            valid_df = path_poly_junction_layer[comp_element_layer_mask]
            if not valid_df.empty:
                layer_jj_to_port.append((comp_name, element, impedance))
        return layer_jj_to_port

    def get_ignored_jjs_in_layer(self, ignored_jjs, layer_num: int) -> list:
        """Reduce the ignored_jjs list and only return the ignored_jjs for layer_nums.

        Args:
            ignored_jjs (_type_): Use definition from render_design.
            layer_num (int): The layer number to render.

        Returns:
            list: Use definition from render_design.
        """
        layer_ignored_jjs = list()
        mask_comp_element = self.path_poly_and_junction_with_valid_comps[
            'layer'] == layer_num
        path_poly_junction_layer = self.path_poly_and_junction_with_valid_comps[
            mask_comp_element]

        for comp_name, element in ignored_jjs:
            id = self.design.components[comp_name].id
            comp_element_layer_mask = (
                path_poly_junction_layer['component'] == id)
            valid_df = path_poly_junction_layer[comp_element_layer_mask]
            if not valid_df.empty:
                layer_ignored_jjs.append((comp_name, element))
        return layer_ignored_jjs

    # yapf: disable
    def render_element(self,
                        qgeom: pd.Series,
                        is_junction: bool,
                        port_list: Union[list, None],
                        jj_to_port: Union[list, None],
                        ignored_jjs: Union[list, None]):
        # yapf: enable
        """Render an individual shape whose properties are listed in a row of
        QGeometry table. Junction elements are handled separately from non-
        junction elements, as the former consist of two rendered shapes, not
        just one.

        Args:
            qgeom (pd.Series): GeoSeries of element properties.
            is_junction (bool): Whether or not qgeom belongs to junction table.
            port_list (Union[list, None], optional): _description_. Defaults to None.
            jj_to_port (Union[list, None], optional): _description_. Defaults to None.
            ignored_jjs (Union[list, None], optional): _description_. Defaults to None.
        """
        qc_shapely = qgeom.geometry
        if is_junction:
            if self.should_render_junction(qgeom, port_list, jj_to_port,
                                           ignored_jjs):
                self.render_element_junction(qgeom, port_list, jj_to_port)
        else:
            if isinstance(qc_shapely, shapely.geometry.Polygon):
                self.render_element_poly(qgeom)
            elif isinstance(qc_shapely, shapely.geometry.LineString):
                self.render_element_path(qgeom)

    def should_render_junction(self, qgeom: pd.Series, port_list: Union[list,
                                                                        None],
                               jj_to_port: Union[list, None],
                               ignored_jjs: Union[list, None]) -> bool:
        """Expected to be replaced by grandchild of this class.

        Args:
            qgeom (pd.Series): _description_
            port_list (Union[list, None]): _description_
            jj_to_port (Union[list, None]): _description_
            ignored_jjs (Union[list, None]): _description_

        Returns:
            bool: _description_
        """
        pass

    def render_element_junction(self,
                                qgeom: pd.Series,
                                port_list: Union[list, None] = None,
                                jj_to_port: Union[list, None] = None):
        """
        Render a Josephson junction consisting of
            1. A rectangle of length pad_gap and width inductor_width. Defines lumped element
               RLC boundary condition.
            2. A line that is later used to calculate the voltage in post-processing analysis.

        At this method, ASSUME that junction is not in the ignored_jjs

        Args:
            qgeom (pd.Series): GeoSeries of element properties.
        """
        #ansys_options = dict(transparency=0.0)

        qc_name = "Lj_" + str(qgeom["component"])
        qc_elt = get_clean_name(qgeom["name"])
        qc_shapely = qgeom.geometry

        # Hardcode in datatype since it is not passed in qgeometry  tables.
        result = self.design.ls.get_properties_for_layer_datatype(
            ['thickness', 'z_coord', 'material', 'fill'],
            qgeom['layer'],
            datatype=0)

        if result:
            thickness, z_coord, material, fill_value = result
        else:
            self.design.ls.layer_stack_handler_pilot_error()

        qc_width = parse_entry(qgeom.width)

        name = f"{qc_name}{QAnsysRenderer.NAME_DELIM}{qc_elt}"

        endpoints = parse_entry(list(qc_shapely.coords))
        endpoints_3d = to_vec3D(endpoints, (z_coord + (thickness / 2))).tolist()
        x0, y0, z0 = endpoints_3d[0]
        x1, y1, z0 = endpoints_3d[1]

        self.assign_mesh.append("JJ_rect_" + name)

        # Draw line
        poly_jj_name = "JJ_" + name + "_"

        poly_jj = self.current_app.modeler.create_polyline(
            endpoints_3d,
            #[endpoints_3d[0], endpoints_3d[1]],
            name=poly_jj_name,
            close_surface=False)
        poly_jj.show_arrow = True
        poly_jj.color = (128, 0, 128)

        jj_name = "JJ_rect_" + name

        poly_sheet = self.current_app.modeler.create_polyline(
            endpoints_3d,
            #[endpoints_3d[0], endpoints_3d[1]],
            name=jj_name,
            close_surface=False,
            xsection_type='Line',
            xsection_width=qc_width)

        solution_type = self.current_app.solution_type

        # If junction not in the port list or ignore list, THEN sheet inductance, if not in jj_to_port.
        # At this method, ASSUME that junction is not in the ignored_jjs
        if solution_type == 'Modal' and self.jj_to_port_is_valid and isinstance(
                jj_to_port, list) and jj_to_port:
            # Only use this logic if user defines drivenmodal solution_type,
            # and  jj_to_port list is provided by user.

            self.jj_to_port_list_dict_populate(jj_to_port)
            search_junction = (self.design._components[qgeom["component"]].name,
                               qgeom['name'])
            if search_junction in self.jj_to_port_dict.keys():
                impedance = self.jj_to_port_dict[search_junction]
                jj_name_port = "JJ_rect_" + "R_" + str(
                    qgeom["component"]) + "_" + str(qgeom['name'])
                jj_to_port_boundary = self.current_app.create_lumped_port_to_sheet(
                    sheet_name=poly_sheet.name,
                    portname=jj_name_port,
                    axisdir=endpoints_3d,
                    impedance=impedance)
        else:
            # Looks QGeometry table to pull Lj and Cj
            jj_geom_table = self.design.qgeometry.tables['junction']
            jj_of_interest = jj_geom_table[jj_geom_table['name'] == qc_elt]
            Lvalue = parse_entry(jj_of_interest['aedt_hfss_inductance'][0])
            Cvalue = parse_entry(jj_of_interest['aedt_hfss_capacitance'][0])

            lumped_rlc_boundary = self.current_app.assign_lumped_rlc_to_sheet(
                sheet_name=poly_sheet.name,
                sourcename=f'rlc_{poly_sheet.name}',
                axisdir=endpoints_3d,
                Lvalue=Lvalue,
                Cvalue=Cvalue)

    def draw_sample_holder(self):
        """Adds a vacuum box to HFSS design.  The xy coordinates are determined by
        results of using box_plus_buffer.  The z coordinate of the box is determined
        by default_options of this class. This the user can update the options prior
        to render_design.

        Note: the height of the box is NOT found by going to Multiplanar Design.
        """
        if "sample_holder_top" in self.design.variables.keys():
            p = self.design.variables
        else:
            p = self.options

        top = parse_entry(p["sample_holder_top"],
                          convert_to_unit=self.current_app.modeler.model_units)
        bottom = parse_entry(
            p["sample_holder_bottom"],
            convert_to_unit=self.current_app.modeler.model_units)
        vac_height = top - bottom
        half_vac_height = (vac_height / 2)
        half_z_coord = top - half_vac_height

        (minx, miny, maxx, maxy) = self.box_for_ansys

        # yapf: disable
        vacuum_rect = [[minx, miny, half_z_coord],
                     [maxx, miny, half_z_coord],
                     [maxx, maxy, half_z_coord],
                     [minx, maxy, half_z_coord],
                     [minx, miny, half_z_coord]]
        # yapf: enable

        box_name = "vacuum_box"

        sample_holder_rectangle = self.current_app.modeler.create_polyline(
            vacuum_rect,
            name=box_name,
            cover_surface=True,
            close_surface=True,
            matname="vacuum")

        self.current_app.modeler.thicken_sheet(sample_holder_rectangle.id,
                                               thickness=vac_height,
                                               bBothSides=True)

    def add_ports(self, port_list: list, port_list_dict: dict, layer_num: int):
        # Should be overwritten by drivenmodal, since eigenmode does not have ports.
        pass

    def add_jj_to_ports(self, jj_to_port: list, jj_to_port_dict: dict,
                        layer_num: int):
        # Should be overwritten by drivenmodal, since eigenmode does not have ports.
        pass


# yapf: disable
    def valid_input_arguments(self,
                            open_pins: Union[list, None],
                            port_list: Union[list, None],
                            jj_to_port: Union[list, None],
                            ignored_jjs: Union[list, None]):
        """The arguments should be valid names in design.  Also (component, pin)
        pairs can not be repeated within the four arguments.

        Args:
            open_pins (Union[list, None]): _description_
            port_list (Union[list, None]): _description_
            jj_to_port (Union[list, None]): _description_
            ignored_jjs (Union[list, None]): _description_

        Returns:
            _type_: _description_
        """
        # yapf: enable

        pair_used = list()

        if open_pins:
            self.open_pins_is_valid = self.confirm_open_pins_are_valid_names(
                open_pins, pair_used)
            if self.open_pins_is_valid == False:
                self.logger.error(f'Arguments are not in Design for open_pins.')
        else:
            self.open_pins_is_valid = True

        if port_list:
            self.port_list_is_valid = self.confirm_port_list_have_valid_request(
                port_list, pair_used)
            if self.port_list_is_valid == False:
                self.logger.error(f'Arguments are not in Design for port_list.')
        else:
            self.port_list_is_valid == True

        if jj_to_port:
            self.jj_to_port_is_valid = self.confirm_jj_to_port_has_valid_request(
                jj_to_port, pair_used)
            if self.jj_to_port_is_valid == False:
                self.logger.error(
                    f'Arguments are not in Design for jj_to_port.')
        else:
            self.jj_to_port_is_valid == True

        if ignored_jjs:
            self.ignored_jjs_is_valid = self.confirm_ignored_jjs_has_valid_request(
                ignored_jjs, pair_used)
            if self.ignored_jjs_is_valid == False:
                self.logger.error(
                    f'Arguments are not in Design for ignored_jjs.')
        else:
            self.ignored_jjs_is_valid = True

        if self.open_pins_is_valid == False or self.port_list_is_valid == False or self.jj_to_port_is_valid == False or self.ignored_jjs_is_valid == False:
            return False
        else:
            return True

    def reset_hfss_arguments(self):
        """Reset the value to None each time render_design is started.
        """
        self.open_pins_is_valid = None
        self.port_list_is_valid = None
        self.jj_to_port_is_valid = None
        self.ignored_jjs_is_valid = None
        self.ignored_jjs = []

    def confirm_jj_to_port_has_valid_request(self, jj_to_port: list,
                                             pair_used: list) -> bool:
        """ The jj_to_port, make sure the component_nam and element type are in design.

        Args:
            jj_to_port (list): List of tuples denoted (component_name, element_name, impedance).
            pair_used (list): Passed to multiple checks to ensure (component_name, pin)
                            are not accidentally reused.

        Returns:
            bool: True if pair in junction table.
                False if  any entry in list not found in design.
        """
        # (component_name, element_name, impedance)
        # are added to the list jj_to_port. For example,
        # ('Q1', 'rect_jj', 70)

        for component_name, element_name, impedance in jj_to_port:
            if self.qcomp_ids:
                # User wants to render a subset, ensure part of selection.
                if not self.design.components[
                        component_name].id in self.qcomp_ids:
                    self.design.logger.error(
                        f'You entered the component_name='
                        f'{self.design.components[component_name].name} and the corresponding'
                        f' name is not selection.'
                        f'The geometries will not be rendered to Ansys.')
                    return False
            junction_table = self.design.qgeometry.tables['junction']
            if self.design.components[component_name]:
                comp_id = self.design.components[component_name].id
            else:
                return False

            mask_comp_element = (junction_table['component'] == comp_id) & (
                junction_table['name'] == element_name)
            result_found_df = junction_table[mask_comp_element]

            if len(result_found_df) == 0:
                return False

            if not (isinstance(impedance, int) or isinstance(impedance, float)):
                self.logger.error(f'Impedance should be either float or int. '
                                  f'You have impedance={impedance}')
                return False
            item = (component_name, element_name)
            if item in pair_used:
                self.logger.error(
                    f'You entered the component_name={component_name} and element_name={element_name} pair twice.  '
                    f'The geometries will not be rendered to Ansys.')
                return False

            pair_used.append(item)

        return True

    def confirm_ignored_jjs_has_valid_request(self, ignored_jjs: list,
                                              pair_used: list) -> bool:
        """Check within the qgeometry junction table to confirm the component_name
        and element_name are paired on a row.

        Args:
            ignored_jjs (list): List of tuples of (component name, element_name)
            pair_used (list): Passed to multiple checks to ensure (component_name, pin)
                            are not accidentally reused.

        Returns:
            bool: True if pair in junction table.
                False if  any entry in list not found in design.
        """

        for item in ignored_jjs:
            component_name, element_name = item
            if self.qcomp_ids:
                # User wants to render a subset, ensure part of selection.
                if not self.design.components[
                        component_name].id in self.qcomp_ids:
                    self.design.logger.error(
                        f'You entered the component_name='
                        f'{self.design.components[component_name].name} and the corresponding'
                        f' name is not selection.'
                        f'The geometries will not be rendered to Ansys.')
                    return False
            junction_table = self.design.qgeometry.tables['junction']
            if self.design.components[component_name]:
                comp_id = self.design.components[component_name].id
            else:
                return False
            mask_comp_element = (junction_table['component'] == comp_id) & (
                junction_table['name'] == element_name)
            result_found_df = junction_table[mask_comp_element]

            if len(result_found_df) == 0:
                return False
            if item in pair_used:
                self.logger.error(
                    f'You entered the component_name={component_name} and element_name={element_name} pair twice.  '
                    f'The geometries will not be rendered to Ansys.')
                return False

            pair_used.append(item)
        return True

    def confirm_port_list_have_valid_request(self, port_list: list,
                                             pair_used: list) -> bool:
        """Check if all names of ports are within design.  Otherwise log an error.

        Args:
            port_list (list): The user input for ports. List of tuples of pins to
                            be rendered as ports.
            pair_used (list): Passed to multiple checks to ensure (component_name, pin)
                            are not accidentally reused.

        Returns:
            bool: True if all ports found in design.
                 False if any one component or pin is not found in design.
        """
        for comp_name, pin, impedance in port_list:
            if self.qcomp_ids:
                # User wants to render a subset, ensure part of selection.
                if not self.design.components[comp_name].id in self.qcomp_ids:
                    self.design.logger.error(
                        f'You entered the component_name='
                        f'{self.design.components[comp_name].name} and the corresponding'
                        f' name is not selection.'
                        f'The geometries will not be rendered to Ansys.')
                    return False
            if comp_name not in self.design.components.keys():
                self.port_list_names_not_valid(comp_name, pin)
                return False
            if pin not in self.design.components[comp_name].pins.keys():
                self.port_list_names_not_valid(comp_name, pin)
                return False
            if not (isinstance(impedance, int) or isinstance(impedance, float)):
                self.logger.error(f'Impedance should be either float or int. '
                                  f'You have impedance={impedance}')
                return False
            search_item = (comp_name, pin)
            if search_item in pair_used:
                self.logger.error(
                    f'You entered the component_name={comp_name} and pin={pin} pair twice.  '
                    f'The geometries will not be rendered to Ansys.')
                return False

            pair_used.append(search_item)

        return True

    def port_list_dict_populate(self, port_list: list):
        """Convert the port_list to a searchable dict.

        Args:
            port_list (list): _description_
        """
        # This should be overwritten by drivenmodal class.
        pass

    def jj_to_port_list_dict_populate(self, jj_to_port: list):
        """Convert jj_to_port to a searchable dict.

        Args:
            jj_to_port (list): _description_
        """
        # This should be overwritten by drivenmodal class.
        pass

    def set_variable(self, name: str, value: str):
        """
        Sets project-level variable in ANSYS.

        Example usage:
                self.set_variable('Lj_1', '10nH')
                self.set_variable('Cj', '0fF')
                self.set_variable('dimensionless', 1/137)

        Args:
            name (str): Name of variable.
            value (str): Amount and unit associated w/ variable.
                
        """
        variable_manager = self.current_app._variable_manager
        variable_manager[name] = value

    ######### Warnings and Errors##################################################

    def port_list_names_not_valid(self, comp_name: str, pin_name: str):
        """Stop and give error if the component name or pin name from port_list
        are not valid within design.

        Args:
            comp_name (str): Name of component which should have an open pin.
            pin_name (str): Name of pin which should be open.
        """
        self.logger.error(
            f'\n The component name or pin name within list'
            f'\nprovided to render_design are not valid.'
            f'\nYou provided: {comp_name} and {pin_name}'
            f'\n Valid components: {self.design.components.keys()}'
            f'\n Valid pins:{self.design.components[comp_name].pins.keys()}'
            f'\n ALL tuples within port_list will not be addressed.')


# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""File contains basic dictionaries.

File contains a class to contain the basic dictionaries.
"""

from copy import deepcopy

from .toolbox_python.attr_dict import Dict


class DefaultMetalOptions(Dict):
    """`DefaultMetalOptions` is the container for the default options used in:

        1. Components - each time a new component is registered (instantiated).
        2. The metal code codebase, in functions such as drawing and in qdesign base

    Args:
        generic (Dict): Dictionary of options.  Defaults to None.
    """

    default_generic = Dict(
        units='mm',  # Units in which all dimensions are converted as floats
        chip=
        'main',  # Default chip used through codebase, when one is not specified

        # Default options for QDesign variables
        qdesign=Dict(variables=Dict(cpw_width='10 um', cpw_gap='6 um')),

        # Used for numpy.round()
        PRECISION=9,

        # Geometric
        geometry=Dict(
            buffer_resolution=16,  # for shapely buffer
            buffer_mitre_limit=5.0,
        ))
    """
    Define the default generic properties
    """

    def __init__(self, generic: Dict = None):
        """The constructor for the `DefaultMetalOptions` class."""
        if not generic:
            generic = deepcopy(self.default_generic)

        # self._generic = deepcopy(generic) # Do we need to save this?
        super().__init__()
        self.update(generic)

    def update_default_options(self, cust_key: str, cust_value=None):
        """Allow instance of class to update the default_options.

        Args:
            cust_key (str): Type of component
            cust_value (object): Value for the given key.  Defaults to None.
        """
        self[cust_key] = cust_value


# Can't really use this until default_draw_substrate.color_plane is resolved.
class DefaultOptionsRenderer():  # pylint: disable=too-few-public-methods
    """`DefaultOptionsRenderer` is the class that encapsulate generic data used
    throughout qiskit metal classes for renderers.

    This class is a skeleton and is expected to be updated when the renderer is updated.

    Args:
        draw_substrate (Dict): This is the dictionary defining the draw substrate parameters.
        bounding_box (Dict): This is the dictionary defining the bounding box parameters.
    """

    # These are potential dicts that could be used for renderers.
    default_bounding_box = Dict(
        draw_bounding_box=[[0, 0], [0, 0], ['0.890mm', '0.900mm']])
    """
    Define the default bounding box.
    """

    default_draw_substrate = Dict(
        draw_substrate={
            'pos_xy': "['0um', '0um']",
            'size': "['8.5mm', '6.5mm', '-0.750mm']",
            'elevation': 0,
            'ground_plane': 'ground_plane',
            'substrate': 'substrate',
            'material': 'silicon',
            # 'color_plane': DEFAULT.colors.ground_main,      #this needs to change
            'transparency_plane': 0,
            'transparency_substrate': 0,
            'wireframe_substrate': False
        })
    """
    Define the default draw substrate
    """

    def __init__(self,
                 draw_substrate: Dict = default_draw_substrate,
                 bounding_box: Dict = default_bounding_box):
        """The constructor for the `DefaultOptionsRenderer` class."""
        #self.logger = logger

        # Do Not edit the class variable
        self.draw_substrate = deepcopy(draw_substrate)
        self.bounding_box = deepcopy(bounding_box)

        # custom default options
        self.default_options = {}
        self.default_options.update(self.draw_substrate)
        self.default_options.update(self.bounding_box)

    # customize the key/value pairs
    def update_default_options(self,
                               cust_key: str = None,
                               cust_value: Dict = None):
        """Allow instance of class to update the default_options.

        Args:
            cust_key (str): Type of component.  Defaults to None.
            cust_value (Dict): The key/value pairs to describe component.  Defaults to None.

        Returns:
            The return value. True for success, False otherwise.
        """
        assert (cust_key is not None
               ), f'ERROR: Need a key, update_default_options has {cust_key}'
        self.default_options[cust_key] = cust_value



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import BaseQubit


class TransmonCross(BaseQubit):  # pylint: disable=invalid-name
    """The base `TransmonCross` class.

    Inherits `BaseQubit` class.

    Simple Metal Transmon Cross object. Creates the X cross-shaped island,
    the "junction" on the south end, and up to 3 connectors on the remaining arms
    (claw or gap).

    'claw_width' and 'claw_gap' define the width/gap of the CPW line that
    makes up the connector. Note, DC SQUID currently represented by single
    inductance sheet

    Add connectors to it using the `connection_pads` dictionary. See BaseQubit for more
    information.

    Sketch:
        Below is a sketch of the qubit
        ::

                                        claw_length
            Claw:       _________                    Gap:
                        |   ________________             _________    ____________
                  ______|  |                             _________|  |____________
                        |  |________________
                        |_________


    .. image::
        TransmonCross.png

    .. meta::
        :description: Transmon Cross

    BaseQubit Default Options:
        * connection_pads: Empty Dict -- The dictionary which contains all active connection lines for the qubit.
        * _default_connection_pads: empty Dict -- The default values for the (if any) connection lines of the qubit.

    Default Options:
        * cross_width: '20um' -- Width of the CPW center trace making up the Crossmon
        * cross_length: '200um' -- Length of one Crossmon arm (from center)
        * cross_gap: '20um' -- Width of the CPW gap making up the Crossmon
        * _default_connection_pads: Dict
            * connector_type: '0' -- 0 = Claw type, 1 = gap type
            * claw_length: '30um' -- Length of the claw 'arms', measured from the connector center trace
            * ground_spacing: '5um' -- Amount of ground plane between the connector and Crossmon arm (minimum should be based on fabrication capabilities)
            * claw_width: '10um' -- The width of the CPW center trace making up the claw/gap connector
            * claw_gap: '6um' -- The gap of the CPW center trace making up the claw/gap connector
            * connector_location: '0' -- 0 => 'west' arm, 90 => 'north' arm, 180 => 'east' arm
    """

    default_options = Dict(
        cross_width='20um',
        cross_length='200um',
        cross_gap='20um',
        chip='main',
        _default_connection_pads=Dict(
            connector_type='0',  # 0 = Claw type, 1 = gap type
            claw_length='30um',
            ground_spacing='5um',
            claw_width='10um',
            claw_gap='6um',
            claw_cpw_length='40um',
            claw_cpw_width='10um',
            connector_location=
            '0'  # 0 => 'west' arm, 90 => 'north' arm, 180 => 'east' arm
        ))
    """Default options."""

    component_metadata = Dict(short_name='Cross',
                              _qgeometry_table_poly='True',
                              _qgeometry_table_junction='True')
    """Component metadata"""

    TOOLTIP = """Simple Metal Transmon Cross."""

    ##############################################MAKE######################################################

    def make(self):
        """This is executed by the GUI/user to generate the qgeometry for the
        component."""
        self.make_pocket()
        self.make_connection_pads()

###################################TRANSMON#############################################################

    def make_pocket(self):
        """Makes a basic Crossmon, 4 arm cross."""

        # self.p allows us to directly access parsed values (string -> numbers) from the user option
        p = self.p

        cross_width = p.cross_width
        cross_length = p.cross_length
        cross_gap = p.cross_gap

        # access to chip name
        chip = p.chip

        # Creates the cross and the etch equivalent.
        cross_line = draw.shapely.ops.unary_union([
            draw.LineString([(0, cross_length), (0, -cross_length)]),
            draw.LineString([(cross_length, 0), (-cross_length, 0)])
        ])

        cross = cross_line.buffer(cross_width / 2, cap_style=2)
        cross_etch = cross.buffer(cross_gap, cap_style=3, join_style=2)

        # The junction/SQUID
        #rect_jj = draw.rectangle(cross_width, cross_gap)
        #rect_jj = draw.translate(rect_jj, 0, -cross_length-cross_gap/2)
        rect_jj = draw.LineString([(0, -cross_length),
                                   (0, -cross_length - cross_gap)])

        #rotate and translate
        polys = [cross, cross_etch, rect_jj]
        polys = draw.rotate(polys, p.orientation, origin=(0, 0))
        polys = draw.translate(polys, p.pos_x, p.pos_y)

        [cross, cross_etch, rect_jj] = polys

        # generate qgeometry
        self.add_qgeometry('poly', dict(cross=cross), chip=chip)
        self.add_qgeometry('poly',
                           dict(cross_etch=cross_etch),
                           subtract=True,
                           chip=chip)
        self.add_qgeometry('junction',
                           dict(rect_jj=rect_jj),
                           width=cross_width,
                           chip=chip)


############################CONNECTORS##################################################################################################

    def make_connection_pads(self):
        """Goes through connector pads and makes each one."""
        for name in self.options.connection_pads:
            self.make_connection_pad(name)

    def make_connection_pad(self, name: str):
        """Makes individual connector pad.

        Args:
            name (str) : Name of the connector pad
        """

        # self.p allows us to directly access parsed values (string -> numbers) from the user option
        p = self.p
        cross_width = p.cross_width
        cross_length = p.cross_length
        cross_gap = p.cross_gap

        # access to chip name
        chip = p.chip

        pc = self.p.connection_pads[name]  # parser on connector options
        c_g = pc.claw_gap
        c_l = pc.claw_length
        c_w = pc.claw_width
        c_c_w = pc.claw_cpw_width
        c_c_l = pc.claw_cpw_length
        g_s = pc.ground_spacing
        con_loc = pc.connector_location

        claw_cpw = draw.box(-c_w, -c_c_w / 2, -c_c_l - c_w, c_c_w / 2)

        if pc.connector_type == 0:  # Claw connector
            t_claw_height = 2*c_g + 2 * c_w + 2*g_s + \
                2*cross_gap + cross_width  # temp value

            claw_base = draw.box(-c_w, -(t_claw_height) / 2, c_l,
                                 t_claw_height / 2)
            claw_subtract = draw.box(0, -t_claw_height / 2 + c_w, c_l,
                                     t_claw_height / 2 - c_w)
            claw_base = claw_base.difference(claw_subtract)

            connector_arm = draw.shapely.ops.unary_union([claw_base, claw_cpw])
            connector_etcher = draw.buffer(connector_arm, c_g)
        else:
            connector_arm = draw.box(0, -c_w / 2, -4 * c_w, c_w / 2)
            connector_etcher = draw.buffer(connector_arm, c_g)

        # Making the pin for  tracking (for easy connect functions).
        # Done here so as to have the same translations and rotations as the connector. Could
        # extract from the connector later, but since allowing different connector types,
        # this seems more straightforward.
        port_line = draw.LineString([(-c_c_l - c_w, -c_c_w / 2),
                                     (-c_c_l - c_w, c_w / 2)])

        claw_rotate = 0
        if con_loc > 135:
            claw_rotate = 180
        elif con_loc > 45:
            claw_rotate = -90

        # Rotates and translates the connector polygons (and temporary port_line)
        polys = [connector_arm, connector_etcher, port_line]
        polys = draw.translate(polys, -(cross_length + cross_gap + g_s + c_g),
                               0)
        polys = draw.rotate(polys, claw_rotate, origin=(0, 0))
        polys = draw.rotate(polys, p.orientation, origin=(0, 0))
        polys = draw.translate(polys, p.pos_x, p.pos_y)
        [connector_arm, connector_etcher, port_line] = polys

        # Generates qgeometry for the connector pads
        self.add_qgeometry('poly', {f'{name}_connector_arm': connector_arm},
                           chip=chip)
        self.add_qgeometry('poly',
                           {f'{name}_connector_etcher': connector_etcher},
                           subtract=True,
                           chip=chip)

        self.add_pin(name, port_line.coords, c_c_w)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# modified by Chalmers/SK 20210611 to add DesignFlipChip
"""
=================================================
Designs (:mod:`qiskit_metal.designs`)
=================================================

.. currentmodule:: qiskit_metal.designs

Module containing all Qiskit Metal designs.

.. _qdesign:

QDesign
---------------

.. autosummary::
    :toctree: .

    QDesign


DesignPlanar
---------------

.. autosummary::
    :toctree: .

    DesignPlanar


MultiPlanar
---------------

.. autosummary::
    :toctree: .

    MultiPlanar


DesignFlipChip
---------------

.. autosummary::
    :toctree: .

    DesignFlipChip

QNet
---------------

.. autosummary::
    :toctree: .

    QNet


InterfaceComponents
-------------------

.. autosummary::
    :toctree: .

    Components
"""

from .. import Dict
from .. import is_design
from .design_base import QDesign
from .design_planar import DesignPlanar
from .design_multiplanar import MultiPlanar
from .design_flipchip import DesignFlipChip
from .net_info import QNet
from .interface_components import Components



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QLabel, QVBoxLayout, QWidget


class QWidget_PlaceholderText:
    """QTableView or Tree with placeholder text if empty.

    This class acts as a mixin for placeholder text functionality.
    """
    __placeholder_text = "The table is empty."

    def __init__(self, placeholder_text: str = None, parent=None):
        """
        Args:
            placeholder_text (str): Placeholder text. Defaults to None.
            parent: Parent widget.
        """
        self._placeholder_text = placeholder_text if placeholder_text else self.__placeholder_text
        self._placeholder_label = QLabel(self._placeholder_text, parent)
        self.setup_placeholder_label(parent)

    def setup_placeholder_label(self, parent):
        """Set up the placeholder label."""
        # Update placeholder text
        self.update_placeholder_text()

        # Ensure layout is present
        if parent and not parent.layout():
            layout = QVBoxLayout(parent)
            parent.setLayout(layout)

        if parent:
            parent.layout().addWidget(self._placeholder_label)

    def update_placeholder_text(self, text=None):
        """Update the placeholder text to the given string.

        Args:
            text (str): New placeholder text. Defaults to None.
        """
        if text:
            self._placeholder_text = text

        label = self._placeholder_label
        label.setText(self._placeholder_text)

        # Styling
        label.setWordWrap(True)
        label.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        label.setAutoFillBackground(False)
        label.setAttribute(Qt.WA_TranslucentBackground)

        # Placeholder text color
        palette = self._placeholder_label.palette()
        color = palette.color(palette.ColorRole.WindowText)  # Correct usage
        palette.setColor(palette.ColorRole.Text, color)
        label.setPalette(palette)

    def show_placeholder_text(self):
        """Show the placeholder text."""
        self._placeholder_label.show()

    def hide_placeholder_text(self):
        """Hide the placeholder text."""
        self._placeholder_label.hide()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './net_list_ui.ui',
# licensing of './net_list_ui.ui' applies.
#
# Created: Wed May 25 09:29:27 2022
#      by: pyside6-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore, QtGui, QtWidgets


class Ui_NetListWindow(object):

    def setupUi(self, NetListWindow):
        NetListWindow.setObjectName("NetListWindow")
        NetListWindow.resize(841, 623)
        self.centralwidget = QtWidgets.QWidget(NetListWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.tableElements = QtWidgets.QTableView(self.centralwidget)
        self.tableElements.setProperty("showDropIndicator", False)
        self.tableElements.setDragDropOverwriteMode(False)
        self.tableElements.setAlternatingRowColors(True)
        self.tableElements.setObjectName("tableElements")
        self.verticalLayout.addWidget(self.tableElements)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        NetListWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar()
        self.menubar.setGeometry(QtCore.QRect(0, 0, 841, 24))
        self.menubar.setObjectName("menubar")
        NetListWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(NetListWindow)
        self.statusbar.setObjectName("statusbar")
        NetListWindow.setStatusBar(self.statusbar)

        self.retranslateUi(NetListWindow)
        QtCore.QMetaObject.connectSlotsByName(NetListWindow)

    def retranslateUi(self, NetListWindow):
        NetListWindow.setWindowTitle(
            QtWidgets.QApplication.translate("NetListWindow", "MainWindow",
                                             None, -1))



from typing import Dict, Any
import sequencing as seq

from qiskit_metal.analyses.quantization.lom_core_analysis import Subsystem


def to_external_system(lom_subsystem: Subsystem, mapping: Dict[str, Any]):
    """Convert a Metal LOM subsystem to an external system based on a custom mapping

    Args:
        lom_subsystem (Subsystem): the LOM subsystem
        mapping (Dict[str, Any]): custom mapping for the conversion,
            where the keys are LOM subsystem types and values are
            the custom external systems (which can be arbitrary types)

    Raises:
        ValueError: throws when the provided LOM subsystem type
            doesn't exist in the custom mapping

    Returns:
        [type]: external system
    """
    if lom_subsystem.sys_type not in mapping:
        raise ValueError(
            f'LOM subsystem of type {lom_subsystem.sys_type} cannot be converted to an extern system.'
        )

    return mapping[lom_subsystem.sys_type]


##-----------------------------------------------------------------------------------
## Add new custom external system mapping or modify or extend existing mappings here
##-----------------------------------------------------------------------------------

LOM_SUBSYSTEM_TO_SEQ_MODE = {
    'TRANSMON': seq.Transmon,
    'FLUXONIUM': seq.Qubit,
    'TL_RESONATOR': seq.Cavity,
    'LUMPED_RESONATOR': seq.Cavity
}



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""GUI front-end interface for Qiskit Metal in PySide2."""
# pylint: disable=invalid-name

import logging
import os
import sys
from copy import deepcopy
from pathlib import Path

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QTimer
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QApplication, QFileDialog, QMainWindow, QMessageBox, QDockWidget

from .. import config
from ..toolbox_python._logging import setup_logger
from . import __version__
from .utility._handle_qt_messages import slot_catch_error
from .widgets.log_widget.log_metal import LogHandler_for_QTextLog


class QMainWindowExtensionBase(QMainWindow):
    """This contains all the functions that the gui needs to call directly from
    the UI.

    Extends the `QMainWindow` class.
    """

    def __init__(self):
        """"""
        super().__init__()
        # Set manually
        self.handler: QMainWindowBaseHandler = None
        self.force_close = False

    @property
    def logger(self) -> logging.Logger:
        """Get the logger."""
        return self.handler.logger

    @property
    def settings(self) -> QtCore.QSettings:
        """Get the settings."""
        return self.handler.settings

    @property
    def gui(self) -> 'QMainWindowBaseHandler':
        """Get the GUI."""
        self.handler

    def _remove_log_handlers(self):
        """Remove the log handlers."""
        if hasattr(self, 'log_text'):
            self.log_text.remove_handlers(self.logger)

    def destroy(self,
                destroyWindow: bool = True,
                destroySubWindows: bool = True):
        """When the window is cleaned up from memory.

        Args:
            destroyWindow (bool): Whether or not to destroy the window.  Defaults to True.
            destroySubWindows (bool): Whether or not to destroy sub windows  Defaults to True.
        """
        self._remove_log_handlers()
        super().destroy(destroyWindow=destroyWindow,
                        destroySubWindows=destroySubWindows)

    def save_window_settings(self):
        """Save the window settings."""
        self.logger.info('Saving window state')
        # get the current size and position of the window as a byte array.
        self.settings.setValue('metal_version', __version__)
        self.settings.setValue('geometry', self.saveGeometry())
        self.settings.setValue('windowState', self.saveState())
        self.settings.setValue('stylesheet', self.handler._stylesheet)

    def restore_window_settings(self):
        """Call a Qt built-in function to restore values from the settings
        file.

        Raises:
            Exception: Error in restoration
        """

        version_settings = self.settings.value('metal_version',
                                               defaultValue='0')
        if __version__ > version_settings:
            self.logger.debug(
                f"Clearing window settings [{version_settings}]...")
            self.settings.clear()

        try:
            self.logger.debug("Restoring window settings...")

            # should probably call .encode("ascii") here
            geom = self.settings.value('geometry', '')
            if isinstance(geom, str):
                geom = geom.encode("ascii")
            self.restoreGeometry(geom)

            window_state = self.settings.value('windowState', '')
            if isinstance(window_state, str):
                window_state = window_state.encode("ascii")
            self.restoreState(window_state)

            self.handler.load_stylesheet(
                self.settings.value('stylesheet',
                                    self.handler._stylesheet_default))

            # TODO: Recent files
        except Exception as e:
            self.logger.error(f'ERROR [restore_window_settings]: {e}')

    def bring_to_top(self):
        """Bring window to top.

        Note that on Windows, this doesn't always work quite well.
        """
        self.raise_()
        self.activateWindow()

    def get_screenshot(self,
                       name='shot',
                       type_='png',
                       display=True,
                       disp_ops=None):
        """Grad a screenshot of the main window, save to file, copy to
        clipboard and visualize in jupyter.

        Args:
            name (string): File name without extension
            type_ (string): File format and name extension
            display (bool): Indicates whether to visualize or not in jupyter notebook
            disp_ops (dict): Used to pass options to IPython.display.Image (example: width)
        """

        path = Path(name + '.' + type_).resolve()

        # grab the main window
        screenshot: QtGui.QPixmap = self.grab()
        screenshot.save(str(path), type_)

        # copy to clipboard
        QtWidgets.QApplication.clipboard().setPixmap(screenshot)
        self.logger.info(
            f'Screenshot copied to clipboard and saved to:\n {path}')

        # visualize in jupyter (adapt resolution and width first)
        if display:
            from IPython.display import Image, display
            _disp_ops = dict(width=500)
            _disp_ops.update(disp_ops or {})
            width_to_scale = round(
                min(_disp_ops['width'] * 1.5, screenshot.width()))
            if not width_to_scale == screenshot.width():
                path = Path(name + str(width_to_scale) + '.' + type_).resolve()
                screenshot = screenshot.scaledToWidth(
                    width_to_scale, mode=QtCore.Qt.SmoothTransformation)
                screenshot.save(str(path), type_)
            display(Image(filename=str(path), **_disp_ops))

    def toggle_all_docks(self, do_hide: bool = None):
        """Show or hide all docks.

        Args:
            do_hide (bool): Hide or show.  Defaults to None -- toggle.
        """
        # Get all docks to show/hide. Ignore edit source
        docks = [
            widget for widget in self.children()
            if isinstance(widget, QDockWidget)
        ]
        docks = list(
            filter(
                lambda x: not x.windowTitle().lower().startswith('edit source'),
                docks))
        docks += [
            widget for widget in self.gui.plot_win.children()
            if isinstance(widget, QDockWidget)
        ]  # specific

        if do_hide is None:
            dock_states = {dock: dock.isVisible() for dock in docks}
            do_hide = any(
                dock_states.values())  # if any are visible then hide all

        for dock in docks:
            if do_hide:
                dock.hide()
            else:
                dock.show()

        # TODO: small -- fix, changes which dock is on top or now

    ##################################################################
    # For actions

    @slot_catch_error()
    def _screenshot(self, _=None):
        """Used to call from action."""
        self.get_screenshot()

    @slot_catch_error()
    def load_stylesheet_default(self, _=None):
        """Used to call from action."""
        self.handler.load_stylesheet('default')

    @slot_catch_error()
    def load_stylesheet_metal_dark(self, _=None):
        """Used to call from action."""
        self.handler.load_stylesheet('metal_dark')

    @slot_catch_error()
    def load_stylesheet_dark(self, _=None):
        """Used to call from action."""
        self.handler.load_stylesheet('qdarkstyle')

    @slot_catch_error()
    def load_stylesheet_open(self, _=None):
        """Used to call from action."""
        default_path = str(self.gui.path_stylesheets)
        filename = QFileDialog.getOpenFileName(
            self, 'Select Qt stylesheet file `.qss`', default_path)[0]
        if filename:
            self.logger.info(f'Attempting to load stylesheet file {filename}')

            self.handler.load_stylesheet(filename)


class QMainWindowBaseHandler():
    """Abstract Class to wrap and handle main window (QMainWindow).

    Assumes a UI that has:
        * log_text: a QText for logging

    Assumes we have functions:
        * setup_logger

    Assumes we have objects:
         * config.log.format
         * config.log.datefmt
         * config._ipython
    """
    _myappid = 'QiskitMetal'
    _img_logo_name = 'my_logo.png'
    _img_folder_name = '_imgs'
    _dock_names = []
    _QMainWindowClass = QMainWindowExtensionBase
    __gui_num__ = -1  # used to count the number of gui instances
    _stylesheet_default = 'default'

    @staticmethod
    def __UI__() -> QMainWindow:  # pylint: disable=invalid-name
        """Abstract, replace with UI class."""
        return None

    def __init__(self, logger: logging.Logger = None, handler=False):
        """Can pass in logger.

        Args:
            logger (logging.Logger): The logger.  Defaults to None.
            handler (bool): Not used.  Defaults to False.

        Attributes:
            _settings: Used to save the state of the window
                This information is often stored in the system
                registry on Windows, and in property list files on macOS and iOS.
        """
        self.__class__.__gui_num__ += 1  # used to give a unique identifier

        self.config = deepcopy(config.GUI_CONFIG)
        self.settings = QtCore.QSettings(self._myappid, 'MainWindow')

        # Logger
        if not logger:
            # print('Setting up logger')
            logger = setup_logger(
                # so that they are not all the same
                f'gui{self.__class__.__gui_num__ }',
                config.log.format,
                config.log.datefmt,
                force_set=True,
                create_stream=self.config.stream_to_std,
            )
            log_level = int(
                getattr(logging, self.config.logger.get('level', 'DEBUG')))
            logger.setLevel(log_level)

        self.logger = logger
        self._log_handler = None  # defined in self._setup_logger
        self._stylesheet = self._stylesheet_default  # set by load_stylesheet

        # File paths
        self.path_gui = self._get_file_path()  # Path to gui folder
        self.path_imgs = Path(self.path_gui) / \
            self._img_folder_name  # Path to gui imgs folder
        if not self.path_imgs.is_dir():
            text = f'Bad File path for images! {self.path_imgs}'
            print(text)
            self.logger.error(text)

        # Main Window and App level
        self.main_window = self._QMainWindowClass()
        self.main_window.handler = self
        self.qApp = self._setup_qApp()
        self._setup_main_window_tray()

        # Style and window
        self._style_sheet_path = None
        self.style_window()

        # UI
        self.ui = self.__UI__()
        self.ui.setupUi(self.main_window)
        self.main_window.ui = self.ui

        self.ui.log_text.dock_window = self.ui.dockLog
        self._setup_logger()
        self._setup_window_size()

        self._ui_adjustments()  # overwrite

        self.main_window.restore_window_settings()

    @property
    def path_stylesheets(self):
        """Returns the path to the stylesheet."""
        return Path(self.path_gui) / 'styles'

    def style_window(self):
        """Styles the window."""
        # fusion macintosh # windows
        self.main_window.setStyle(QtWidgets.QStyleFactory.create("fusion"))
        # TODO; add stlyesheet to load here - maybe pull form settings

    def _ui_adjustments(self):
        """Any touchups to the loaded ui that need be done soon."""
        pass

    def _get_file_path(self):
        """Get the dir name of the current path in which this file is
        stored."""
        return os.path.dirname(__file__)

    def _setup_qApp(self):
        """Only one qApp can exist at a time, so check before creating one.

        Returns:
            QApplication: a setup QApplication

        There are three classes:
            * QCoreApplication - base class. Used in command line applications.
            * QGuiApplication - base class + GUI capabilities. Used in QML applications.
            * QApplication - base class + GUI + support for widgets. Use it in QtWidgets applications.

        See:
            * https://forum.qt.io/topic/94834/differences-between-qapplication-qguiappication-qcoreapplication-classes/3
            * https://github.com/matplotlib/matplotlib/blob/9984f9c4db7bfb02ffca53b7823acb8f8e223f6a/lib/matplotlib/backends/backend_qt5.py#L98
        """

        self.qApp = QApplication.instance()

        if self.qApp is None:
            self.logger.warning("QApplication.instance is None.")

            # Kickstart()

            # Did it work now
            self.qApp = QApplication.instance()

            if self.qApp is None:

                self.logger.error(r"""ERROR: QApplication.instance is None.
                Did you run a cell with the magic in IPython?
                ```python
                    %gui qt
                ```
                This command allows IPython to integrate itself with the Qt event loop,
                so you can use both a GUI and an interactive prompt together.
                Reference: https://ipython.readthedocs.io/en/stable/config/eventloops.html
                """)

        # for window platofrms only
        # QApplication.platformName() -- on mac: 'cocoa'
        if os.name.startswith('nt'):
            # Arbitrary string, needed for icon in taskbar to be custom set proper
            # https://stackoverflow.com/questions/1551605/how-to-set-applications-taskbar-icon-in-windows-7
            import ctypes
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(
                self._myappid)

        app = self.qApp
        if app:
            app.setOrganizationName(r"IBM Quantum Metal")
            app.setOrganizationDomain(r"https://www.ibm.com/quantum-computing")
            app.setApplicationName(r"Qiskit Metal")

        return self.qApp

    def _setup_main_window_tray(self):
        """Sets up the main window tray."""

        if self.path_imgs.is_dir():
            icon = QIcon(str(self.path_imgs / self._img_logo_name))
            self.main_window.setWindowIcon(icon)

            # not sure if this works, but let's try
            self._icon_tray = QtWidgets.QSystemTrayIcon(icon, self.main_window)
            self._icon_tray.show()
            self.qApp.setWindowIcon(icon)

    def create_log_handler(self, name_toshow: str, logger: logging.Logger):
        """Creates a log handler.

        Args:
            name_toshow (str): Display name
            logger (logging.Logger): The logger

        Returns:
            LogHandler_for_QTextLog: A LogHandler_for_QTextLog
        """
        return LogHandler_for_QTextLog(name_toshow, self, self.ui.log_text,
                                       logger)

    @slot_catch_error()
    def _setup_logger(self):
        """Setup logging UI.

        Show wellcome message.
        """
        if hasattr(self.ui, 'log_text'):

            self.ui.log_text.img_path = self.path_imgs

            self._log_handler = self.create_log_handler('GUI', self.logger)

            QTimer.singleShot(1500, self.ui.log_text.welcome_message)

        else:
            self.logger.warning('UI does not have `log_text`')

    def _setup_window_size(self):
        """Setup the window size."""
        if self.config.main_window.auto_size:
            screen = self.qApp.primaryScreen()
            # screen.name()

            rect = screen.availableGeometry()
            rect.setWidth(rect.width() * 0.9)
            rect.setHeight(rect.height() * 0.9)
            rect.setLeft(rect.left() + rect.width() * 0.1)
            rect.setTop(rect.top() + rect.height() * 0.1)
            self.main_window.setGeometry(rect)

    def load_stylesheet(self, path=None):
        """Load and set stylesheet for the main gui.

        Args:
            path (str) : Path to stylesheet or its name.
                Can be: 'default', 'qdarkstyle' or None.
                `qdarkstyle` requires
                >>> pip install qdarkstyle

        Returns:
            bool: False if failure, otherwise nothing

        Raises:
            ImportError: Import failure
        """
        result = True
        if path == 'default' or path is None:
            self._style_sheet_path = 'default'
            self.main_window.setStyleSheet('default')

        elif path == 'qdarkstyle':
            try:
                import qdarkstyle
            except ImportError:
                QMessageBox.warning(
                    self.main_window, 'Failed.',
                    'Error, you did not seem to have installed qdarkstyle.\n'
                    'Please do so from the terminal using\n'
                    ' >>> pip install qdarkstyle')

            os.environ['QT_API'] = 'pyside6'
            self.main_window.setStyleSheet(qdarkstyle.load_stylesheet())

        elif path == 'metal_dark':
            path_full = self.path_stylesheets / 'metal_dark' / 'style.qss'
            # print(f'path_full = {path_full}')
            self._load_stylesheet_from_file(path_full)

        else:
            self._load_stylesheet_from_file(path)

        if result:  # Set successfuly
            self._stylesheet = path
            self.settings.setValue('stylesheet', self._stylesheet)
        else:  # Failed to set
            return False

    def _load_stylesheet_from_file(self, path: str):
        """Load the sylesheet from a file.

        Args:
            path (str): Path to file

        Returns:
            bool: False if failure, otherwise nothing

        Raises:
            Exception: Stylesheet load failure
        """
        # print(f'path = {path}')
        try:
            path = Path(str(path))
            if path.is_file():
                self._style_sheet_path = str(path)
                stylesheet = path.read_text()
                stylesheet = stylesheet.replace(':/metal-styles',
                                                str(self.path_stylesheets))

                # if windows, double the slashes in the paths
                if os.name.startswith('nt'):
                    stylesheet = stylesheet.replace("\\", "\\\\")

                self.main_window.setStyleSheet(stylesheet)
                return True

            else:
                self.logger.error(
                    'Could not find the stylesheet file where expected %s',
                    path)
                return False
        except Exception as e:
            self.logger.error(f'_load_stylesheet_from_file error: {e}')

    def screenshot(self, name='shot', type_='png', display=True, disp_ops=None):
        """Alias for get_screenshot()."""
        self.main_window.get_screenshot(name, type_, display, disp_ops)

    def save_file(self):
        """Save file. Called on exit.

        Raises:
            NotImplementedError: Function not written
        """
        raise NotImplementedError()

    def show(self):
        """Show the main window."""
        self.main_window.show()

    def clear_settings(self):
        """Clear the settings that get saved each time the main window is
        closed.

        This will reset the window layout to the default.
        """
        self.settings.clear()

    def set_font_size(self, font_size: int):
        """Set font size of the applicaiton globally in points.

        Args:
            font_size (int): New font size
        """
        app = self.qApp
        # TODO: DO not just overwrite, but append or update style sheet.
        # Maybe allow user to edit as a whole in a widget
        app.setStyleSheet("QWidget{font-size:" + f"{font_size}" + "pt;}")


def kick_start_qApp():
    """Kick start the application.

    Returns:
        QtCore.QCoreApplication: the application

    Raises:
        AttributeError: Attribute only exists for Qt >= 5.6
        Exception: Magic method failure
    """
    qApp = QtCore.QCoreApplication.instance()

    if qApp is None:
        try:
            # TODO: See
            QtWidgets.QApplication.setAttribute(
                QtCore.Qt.AA_EnableHighDpiScaling)
        except AttributeError:  # Attribute only exists for Qt >= 5.6
            pass

        qApp = QApplication(sys.argv)

        if qApp is None:
            logging.error("QApplication.instance is None.")

            if config._ipython:
                # iPython has magic for loop
                logging.error(
                    "QApplication.instance: Attempt magic IPython %%gui qt5")
                try:
                    from IPython import get_ipython
                    ipython = get_ipython()
                    ipython.run_line_magic("gui", "qt6")

                except Exception as e:
                    print("exception")
                    logging.error(f"FAILED: {e}")
                    print(e)

            else:
                # We are not running form IPython, manually boot
                logging.error(
                    "QApplication.instance: Attempt to manually create qt5 QApplication"
                )
                qApp = QtWidgets.QApplication(["qiskit-metal"])
                qApp.lastWindowClosed.connect(qApp.quit)

    return qApp



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
"""File contains utility functions for check for object types."""

from . import Dict


def is_design(obj):
    """Check if an object is a Metal Design, i.e., an instance of `QDesign`.

    The problem is that the `isinstance` built-in method fails
    when this module is reloaded.

    Args:
        obj (object): Test this object

    Returns:
        bool (bool): True if is a Metal design
    """
    if isinstance(obj, Dict):
        return False

    return hasattr(obj, '__i_am_design__')


def is_component(obj):
    """Check if an object is an instance of `QComponent`.

    The problem is that the `isinstance` built-in method fails
    when this module is reloaded.

    Args:
        obj (object): Test this object

    Returns:
        bool (bool): True if is a Metal object
    """
    if isinstance(obj, Dict):
        return False

    return hasattr(obj, '__i_am_component__')



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
===================================================
Toolbox_python (:mod:`qiskit_metal.toolbox_python`)
===================================================

.. currentmodule:: qiskit_metal.toolbox_python

This folder contains code that is general and not Metal specific.

Headings
---------------

.. autosummary::
    :toctree: .

    MetalTutorialMagics
    Headings
    Color


Submodules
---------------

.. autosummary::
    :toctree: .

    _logging
    display
    utility_functions

"""

from .attr_dict import Dict
from .display import format_dict_ala_z, Headings
from .display import MetalTutorialMagics
from .display import Color



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Models the transmon qubit in the cooper-pair charge basis
and calculate the exact (analytic) solutions

@author: Nick Lanzillo (IBM)
"""
# pylint: disable=invalid-name

import numpy as np
from scipy.special import mathieu_a


class Hcpb_analytic:
    """
    Analytic version of Hamiltonian-model Cooper pair box (Hcpb) class.

    Used calculate the exact eigenvalues for arbitrary Ej, Ec, ng values.
    """

    def __init__(self, Ej: float = None, Ec: float = None, ng: float = 0.5):
        """
        Generate a Cooper-pair box (CPB) model.

        Arguments:
            Ej (float): Josephson energy of the JJ
            Ec (float): Charging energy of the CPB
            ng (float): Offset charge of the CPB (ng=0.5 is the sweet spot).
                        `ng` only needs to run between -0.5 and 0.5.
                        `ng` is defined in units of cooper pairs (2e)
        """

        self._Ej = Ej
        self._Ec = Ec
        self._ng = ng
        self.evals = None
        self.evecs = None
        # Generate the diagonal and offdiagonal components of the Hamiltonian
        #self._gen_operators()
        # compute the eigenvectors and eigenvalues of the CPB
        # all properties can be derived from these
        #self._calc_H()

    def evalue_k(self, k: int):
        """
        Return the eigenvalue of the Hamiltonian for level k.

        Arguments:
            k (int): Index of the eigenvalue

        Returns:
            float: eigenvalue of the Hamiltonian
        """
        if self._ng == 0:
            index = k + 1.0 - ((k + 1.0) % 2.0)
        else:
            index = k + 1.0 - ((k + 1.0) % 2.0) + 2.0 * self._ng * (
                (-1.0)**(k - 0.5 * (np.sign(self._ng) - 1.0)))

        self.evals = self._Ec * mathieu_a(index, -0.5 * self._Ej / self._Ec)
        #return self.evals[k]
        return self.evals



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Tree view for Param Entry Window
"""

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QModelIndex, Signal
from PySide6.QtWidgets import QTreeView, QWidget

from qiskit_metal._gui.widgets.qlibrary_display.proxy_model_qlibrary import LibraryFileProxyModel
from qiskit_metal.toolbox_metal.exceptions import QLibraryGUIException


class TreeViewQLibrary(QTreeView):
    """Handles editing a QComponent

    This class extend the `QTreeView`
    """

    qlibrary_filepath_signal = Signal(str)

    def __init__(self, parent: QWidget):
        """
        Inits TreeViewQLibrary
        Args:
            parent (QtWidgets.QWidget): parent widget
        """
        QTreeView.__init__(self, parent)
        self.tool_tip_str = "Library of QComponents.\nClick one to create it."
        self.expanded.connect(self.onExpanded)

    def setModel(self, model: QtCore.QAbstractItemModel):
        """Overriding setModel to ensure only correct QAbstractItemModel subclass is used

        Args:
            model (QtCore.QAbstractItemModel): Model to be set

        Raises:
            Exception: QLibraryGUIException if model is not LibraryFileProxyModel

        """
        if not isinstance(model, LibraryFileProxyModel):
            raise QLibraryGUIException(
                f"Invalid model. Expected type {LibraryFileProxyModel} but got type {type(model)}"
            )

        super().setModel(model)

    def mousePressEvent(self, event: QtGui.QMouseEvent):
        """Overrides inherited mousePressEvent to emit appropriate filepath signals
         based on which columns were clicked, and to allow user to clear any selections
        by clicking off the displayed tree.

        Args:
            event (QtGui.QMouseEvent): QMouseEvent triggered by user
        """

        index = self.indexAt(event.pos())

        if index.row() == -1:
            self.clearSelection()
            self.setCurrentIndex(QModelIndex())
            return super().mousePressEvent(event)

        model = self.model()
        source_model = self.model().sourceModel()
        full_path = source_model.filePath(model.mapToSource(index))

        if index.column() == source_model.FILENAME:
            if not source_model.isDir(model.mapToSource(index)):
                qis_abs_path = full_path[full_path.
                                         index(__name__.split('.')[0]):]
                self.qlibrary_filepath_signal.emit(qis_abs_path)

        return super().mousePressEvent(event)

    def setToolTip(self, qcomp_tooltip: str):
        """
        Sets tooltip

        Args:
            qcomp_tooltip (str): Tooltip to be set

        """
        if qcomp_tooltip is None or len(qcomp_tooltip) < 1:
            super().setToolTip(self.tool_tip_str)
        else:
            super().setToolTip(qcomp_tooltip)

    def onExpanded(self, index: QModelIndex):
        """Ensure that when we expand the first column names fit in

        Args:
            index (QModelIndex): [description]
        """
        self.resizeColumnToContents(0)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from qiskit_metal.designs import QDesign  # pylint: disable=unused-import
from ..core import QAnalysis
from ..simulation import EigenmodeSim

from ... import Dict


# TODO: eliminate every reference to "renderer" in this file
#  then change inheritance from QSimulation to QAnalysis
class EPRanalysis(QAnalysis):
    """From an input eigenmode dataset, apply the Energy Participation Ratio analysis method.

    Default Setup:
        * junctions (Dict):
            * keys (str): Name of each Non-linear (Josephson) junction to consider in EPR
            * values (Dict):
                * Lj_variable (str): Name of renderer variable that specifies junction inductance.
                * Cj_variable (str): Name of renderer variable that specifies junction capacitance.
                * rect (str): Name of renderer rectangle on which the lumped boundary condition
                    is defined.
                * line (str): Name of renderer line spanning the length of rect
                    (voltage, orientation, ZPF).
        * dissipatives (Dict):
            * keys (str): Categories of dissipattives that can appear in the system. Possible keys
                are: 'dielectrics_bulk', 'dielectric_surfaces', 'resistive_surfaces', 'seams'.
            * values (list of str): Names of the shapes composing that dissipative.
        * cos_trunc (int): truncation of the cosine function
        * fock_trunc (int): truncation of the fock
        * sweep_variable (str): Variable to sweep during EPR

    Data Labels:
        * energy_elec (float): Name given to the current sweep.
        * energy_mag (float): Impedance matrix.
        * energy_elec_sub (float): Admittance matrix.

    """
    default_setup = Dict(junctions=Dict(
        jj=Dict(Lj_variable='Lj', Cj_variable='Cj', rect='', line='')),
                         dissipatives=Dict(dielectrics_bulk=['main']),
                         cos_trunc=8,
                         fock_trunc=7,
                         sweep_variable='Lj')
    """Default setup."""

    # supported labels for data generated from the simulation
    data_labels = ['energy_elec', 'energy_mag', 'energy_elec_sub']
    """Default data labels."""

    # TODO: renderer_name should default to None. Need to create a set renderer method
    def __init__(self, design: 'QDesign' = None, renderer_name: str = None):
        """Performs Energy Participation Ratio (EPR) analysis on a simulated or
        user-provided eigenmode matrix.

        Args:
            design (QDesign): Pointer to the main qiskit-metal design.
                Used to access the QRenderer. Defaults to None.
            renderer_name (str, optional): Which renderer to use. Valid entries: 'hfss'.
                Defaults to None.
        """
        # QAnalysis are expected to either run simulation or use pre-saved sim outputs
        # we use a Dict() to store the sim outputs previously saved. Its key names need
        # to match those found in the correspondent simulation class.
        self.sim = Dict() if renderer_name is None else EigenmodeSim(
            design, renderer_name)
        super().__init__()

    @property
    def energy_elec(self) -> float:
        """Getter

        Returns:
            float: Electric field energy stored in the system based on the eigenmode results.
        """
        return self.get_data('energy_elec')

    @energy_elec.setter
    def energy_elec(self, data: float):
        """Setter

        Args:
            data (float): Electric field energy stored in the system based on the eigenmode results.
        """
        if not isinstance(data, float):
            self.logger.warning(
                'Unsupported type %s. Only accepts float. Please try again.',
                {type(data)})
            return
        self.set_data('energy_elec', data)

    @property
    def energy_mag(self) -> float:
        """Getter

        Returns:
            float: Magnetic field energy stored in the system based on the eigenmode results.
        """
        return self.get_data('energy_mag')

    @energy_mag.setter
    def energy_mag(self, data: float):
        """Setter

        Args:
            data (float): Magnetic field energy stored in the system based on the eigenmode results.
        """
        if not isinstance(data, float):
            self.logger.warning(
                'Unsupported type %s. Only accepts float. Please try again.',
                {type(data)})
            return
        self.set_data('energy_mag', data)

    @property
    def energy_elec_sub(self) -> float:
        """Getter

        Returns:
            float: Electric field energy stored in the substrate based on the eigenmode results.
        """
        return self.get_data('energy_elec_sub')

    @energy_elec_sub.setter
    def energy_elec_sub(self, data: float):
        """Setter

        Args:
            data (float): Electric field energy stored in the substrate based
                on the eigenmode results.
        """
        if not isinstance(data, float):
            self.logger.warning(
                'Unsupported type %s. Only accepts float. Please try again.',
                {type(data)})
            return
        self.set_data('energy_elec_sub', data)

    def run(self, *args, **kwargs):
        """Executes sequentially the system capacitance simulation and lom extraction executing
        the methods EigenmodeSim.run_sim(`*args`, `**kwargs`) and EPRanalysis.run_epr().
        For input parameter, see documentation for EigenmodeSim.run_sim().

        Returns:
            (dict): Pass numbers (keys) and respective energy participation ratio (values).
        """
        if isinstance(self.sim, EigenmodeSim):
            self.sim.run(*args, **kwargs)
        return self.run_epr()

    def run_epr(self, no_junctions=False):
        """Executes the epr analysis from the extracted eigenmode,
        and based on the setup values.
        """
        # wipe data from the previous run (if any)
        self.clear_data()

        self.get_stored_energy(no_junctions)
        if not no_junctions:
            self.run_analysis()
            self.spectrum_analysis(self.setup.cos_trunc, self.setup.fock_trunc)
            try:
                self.report_hamiltonian(self.setup.sweep_variable)
            except AttributeError:
                self.logger.error(
                    "Please install a more recent version of pyEPR (>=0.8.5.3)")

    # TODO: all the epr methods should not use the renderer. Now they are forced to because of the
    #  pyEPR dependency from pinfo. pinfo however is Ansys specific and cannot be generalized as-is
    #  Therefore we need to eliminate pyEPR dependency on pinfo, or re-implement in qiskit-metal

    def epr_start(self, no_junctions=False):
        """
        Initialize epr package.
        """
        # pandas cannot handle Dict so need to convert Dict to dict
        system = dict()
        s = self.setup
        system['junctions'] = {} if no_junctions else {
            k: dict(v) for (k, v) in s.junctions.items()
        }
        system['dissipatives'] = dict(s.dissipatives)
        self.sim.renderer.epr_start(**system)
        return system

    def get_stored_energy(self, no_junctions=False):
        """Calculate the energy stored in the system based on the eigenmode results.
        """
        # execute EPR and energy extraction
        self.energy_elec, self.energy_elec_sub, self.energy_mag = \
            self.sim.renderer.epr_get_stored_energy(**self.epr_start(no_junctions))

        # present a human-friendly output
        print(f"""
        energy_elec_all       = {self.energy_elec}
        energy_elec_substrate = {self.energy_elec_sub}
        EPR of substrate = {self.energy_elec_sub / self.energy_elec * 100 :.1f}%

        energy_mag    = {self.energy_mag}
        energy_mag % of energy_elec_all  = {self.energy_mag / self.energy_elec * 100 :.1f}%
        """)

    def run_analysis(self):
        """Short-cut to the same-name method found in renderers.ansys_renderer.py.
        Eventually, the analysis code needs to be only here, and the renderer method deprecated.
        """
        self.sim.renderer.epr_run_analysis()

    def spectrum_analysis(self, cos_trunc: int = 8, fock_trunc: int = 7):
        """Short-cut to the same-name method found in renderers.ansys_renderer.py.
        Eventually, the analysis code needs to be only here, and the renderer method deprecated.
        """
        self.sim.renderer.epr_spectrum_analysis(cos_trunc, fock_trunc)

    def report_hamiltonian(self, sweep_variable, numeric=True):
        """Short-cut to the same-name method found in renderers.ansys_renderer.py.
        Eventually, the analysis code needs to be only here, and the renderer method deprecated.
        """
        self.sim.renderer.epr_report_hamiltonian(sweep_variable, numeric)

    def get_frequencies(self):
        """Short-cut to the same-name method found in renderers.ansys_renderer.py.
        Eventually, the analysis code needs to be only here, and the renderer method deprecated.
        """
        system = self.epr_start(no_junctions=True)
        return self.sim.renderer.epr_get_frequencies(**system)

    def del_junction(self, name_junction='jj'):
        """Remove a junction from the dictionary setup.junctions

        Args:
            name_junction (str, optional): name of the junction to remove. Defaults to 'jj'.
        """
        if name_junction in self.setup.junctions:
            del self.setup.junctions[name_junction]

    def add_junction(self,
                     name_junction="jj",
                     lj_var="Lj",
                     cj_var='Cj',
                     rect='',
                     line=''):
        """Add a new junction for the EPR analysis

        Args:
            name_junction (str, optional): name of the junction. Defaults to "jj".
            Lj_var (str, optional): Name of the simulator variable referring to
                the inductance. Defaults to "Lj".
            Cj_var (str, optional): Name of the simulator variable referring to
                the capacitance. Defaults to 'Cj'.
            rect (str, optional): Name of the rectangle representing the junction
                in the simulation, as defined during rendering. Defaults to ''.
            line (str, optional): Name of the line representing the junction
                current flow in the simulation, as defined during rendering. Defaults to ''.
        """
        j_dic = self.setup.junctions
        if name_junction in j_dic:
            self.logger.warning(
                f"junction already defined. Overwriting {name_junction}")

        j_dic[name_junction] = Dict({
            'Lj_variable': lj_var,
            'Cj_variable': cj_var,
            'rect': rect,
            'line': line
        })

    def load_simulation_data(self, data_name: str, data):
        """Load simulation data for the following analysis. This will override any data found

        Args:
            data_name (str): name of the variable
            data (Any): simulation output
        """
        self.sim[data_name] = data



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# /*##########################################################################
#
# Copyright (c) https://github.com/benjimin/descartes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Paths and patches"""

from matplotlib.patches import PathPatch
from matplotlib.path import Path
from numpy import asarray, concatenate, ones


class Polygon(object):
    """Adapt Shapely or GeoJSON/geo_interface polygons to a common interface"""

    def __init__(self, context):
        if hasattr(context, 'interiors'):
            self.context = context
        else:
            self.context = getattr(context, '__geo_interface__', context)

    @property
    def geom_type(self):
        return (getattr(self.context, 'geom_type', None) or
                self.context['type'])

    @property
    def exterior(self):
        return (getattr(self.context, 'exterior', None) or
                self.context['coordinates'][0])

    @property
    def interiors(self):
        value = getattr(self.context, 'interiors', None)
        if value is None:
            value = self.context['coordinates'][1:]
        return value


def PolygonPath(polygon: Polygon):
    """Constructs a compound matplotlib path from a Shapely or GeoJSON-like
    geometric object

    Args:
        polygon (Polygon): polygon

    Returns:
        Path: compound matplotlib path
    """
    this = Polygon(polygon)
    assert this.geom_type == 'Polygon'

    def coding(ob):
        # The codes will be all "LINETO" commands, except for "MOVETO"s at the
        # beginning of each subpath
        n = len(getattr(ob, 'coords', None) or ob)
        vals = ones(n, dtype=Path.code_type) * Path.LINETO
        vals[0] = Path.MOVETO
        return vals

    vertices = concatenate([asarray(this.exterior.coords)[:, :2]] +
                           [asarray(r.coords)[:, :2] for r in this.interiors])
    codes = concatenate([coding(this.exterior)] +
                        [coding(r) for r in this.interiors])
    return Path(vertices, codes)


def PolygonPatch(polygon: Polygon, **kwargs):
    """Constructs a matplotlib patch from a geometric object
    
    The `polygon` may be a Shapely or GeoJSON-like object with or without holes.
    The `kwargs` are those supported by the matplotlib.patches.Polygon class
    constructor. Returns an instance of matplotlib.patches.PathPatch.
    Example (using Shapely Point and a matplotlib axes):
      >>> b = Point(0, 0).buffer(1.0)
      >>> patch = PolygonPatch(b, fc='blue', ec='blue', alpha=0.5)
      >>> axis.add_patch(patch)

    Args:
        polygon (Polygon): polygon
        kwargs (dict): dictionary

    Returns:
        PathPatch: matplotlib patch
    """
    return PathPatch(PolygonPath(polygon), **kwargs)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from PySide6.QtCore import QModelIndex
from PySide6.QtWidgets import QDialog, QMainWindow

from .add_delete_table_ui import Ui_MainWindow
from .dialog_popup_ui import Ui_Dialog
from .prop_val_table_model import PropValTable


class PropertyTableWidget(QMainWindow):
    """GUI for variables table with 3 columns: Property, Value, Number.

    Extends the `QMainWindow` class.
    """

    def __init__(self, parent, design=None, gui=None):
        """
        Args:
            parent (QMainWindowExtension): Parent window
            design (QDesign): design.  Defaults to None.
            gui (MetalGUI): the GUI.  Defaults to None.
        """
        super().__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        #
        self._design = design

        # Table View:
        self.table = self.ui.tableView
        self.table.setAlternatingRowColors(True)
        self.table.horizontalHeader().setVisible(True)
        self.table.verticalHeader().setVisible(True)

        # Table Model:
        self.model = PropValTable(design, gui, self.table)
        self.table.setModel(self.model)

        # Display:
        self.show()

    def set_design(self, design):
        """Swap out reference to design, which changes the reference to the
        dictionary.

        Args:
            design (QDesign): The design
        """
        self._design = design
        self.model.set_design(design)

    @property
    def design(self):
        """Returns the design."""
        return self._design

    @property
    def _data(self) -> dict:
        """Returns the variables."""
        if self._design:
            return self._design.variables

    def addRow(self):
        """Add a new row."""
        db = QDialog(self)
        dialog = Ui_Dialog()
        dialog.setupUi(db)
        db.exec()
        key = dialog.key_box.text()
        val = dialog.value_box.text()
        if key and val:
            self.model.add_row(key, val)

    def deleteRow(self):
        """Delete all rows corresponding to selected cells."""
        rowidxlst = list(set(elt.row() for elt in self.table.selectedIndexes()))
        for idx in sorted(rowidxlst, reverse=True):
            self.model.removeRows(idx, 1, QModelIndex())



### BASIC PROGRAM THAT LOOKS FOR OVERLAP BETWEEN QCOMPONENTS
from numpy import size
import geopandas as gpd
import pandas as pd


class QDesignCheck():
    """QDesign_Check contains various design checks, such as
    testing designs in qiskit metal for unintended overlap
    between components and/or connections between components.
    """

    def __init__(self, design: 'QDesign'):
        self.design = design

    def update_design(self, design: 'QDesign'):
        self.design = design

    def overlap_tester(self):
        design = self
        """This particular function tests for overlap amongst qcomponents
        and CPWs. It will catch qubit/qubit overlap, qubit, CPW overlap
        and CPW/CPW overlap.
        """

        for unique_int in self.design._components:
            print(" ")
            print("Component ID:")
            print(unique_int)
            comp = unique_int

            # Get the GeoSeries tables separately for polys and paths, then combine.
            poly = self.design._components[unique_int].qgeometry_table('poly')
            path = self.design._components[unique_int].qgeometry_table('path')
            combined = gpd.GeoDataFrame(
                pd.concat([poly, path], ignore_index=True))
            combined_geo = combined["geometry"]

            # loop within a loop to calculate distance between components
            for unique_int in self.design._components:

                main_counter = 0.0

                # don't calculate distance between a component and itself
                if unique_int == comp:
                    #print("Don't check for collisions between component and itself.")
                    pass
                else:
                    poly_inner = self.design._components[
                        unique_int].qgeometry_table('poly')
                    path_inner = self.design._components[
                        unique_int].qgeometry_table('path')
                    combined_inner = gpd.GeoDataFrame(
                        pd.concat([poly_inner, path_inner], ignore_index=True))
                    combined_geo_inner = combined_inner["geometry"]

                    collision_counter_outer = 0.0

                    n = int(size(combined_geo_inner))
                    for i in range(0, n):

                        #print("inner element:", i)
                        el_inner = combined_geo.crosses(combined_geo_inner[i])

                        m = int(size(el_inner))
                        collision_counter = 0.0
                        for j in range(0, m):
                            if el_inner[j] == 1:
                                collision_counter = collision_counter + 1.0
                            else:
                                pass
                        if collision_counter > 0.0:
                            collision_counter_outer = collision_counter_outer + 1.0
                        else:
                            pass

                    if collision_counter_outer > 0.0:
                        main_counter = main_counter + 1.0
                    else:
                        pass

                if main_counter > 0.0:
                    print("Has a collision with the following QComponent:",
                          unique_int)
                else:
                    pass
